try:
    import nltk
except:
    print("(!) Code in `autogoal.contrib.nltk` requires `nltk`.")
    print("(!) You can install it with `pip install autogoal[nltk]`.")
    raise

from autogoal.contrib.nltk._generated import *
from autogoal.contrib.nltk._manual import *

import os
from pathlib import Path


CONTRIB_NAME = "nltk"
DATA_PATH = Path.home() / ".autogoal" / "contrib" / CONTRIB_NAME / "data"


if DATA_PATH not in nltk.data.path:
    nltk.data.path.insert(0, str(DATA_PATH))


def download():
    os.makedirs(DATA_PATH, exist_ok=True)
    return nltk.download(
        info_or_id=["wordnet", "sentiwordnet", "averaged_perceptron_tagger"],
        download_dir=DATA_PATH,
    )


def status():
    from autogoal.contrib import ContribStatus

    try:
        from nltk.corpus import wordnet
        from nltk.corpus import sentiwordnet

        from nltk.tag import PerceptronTagger

        tagger = PerceptronTagger()
    except LookupError:
        return ContribStatus.RequiresDownload

    return ContribStatus.Ready



from logging import log
import multiprocessing
import warnings
import psutil
import signal
import os
import traceback
import logging
import platform

if platform.system() == "Linux":
    import resource

from autogoal.utils import Mb

logger = logging.getLogger("autogoal")


class RestrictedWorker:
    def __init__(self, function, timeout: int, memory: int):
        self.function = function
        self.timeout = timeout
        self.memory = memory
        signal.signal(signal.SIGXCPU, alarm_handler)

    def _restrict(self):
        if platform.system() == "Linux":
            msoft, mhard = resource.getrlimit(resource.RLIMIT_DATA)
            csoft, chard = resource.getrlimit(resource.RLIMIT_CPU)
            used_memory = self.get_used_memory()

            if self.memory and self.memory > (used_memory + 500 * Mb):
                # memory may be restricted
                resource.setrlimit(resource.RLIMIT_DATA, (self.memory, mhard))
            else:
                warnings.warn("Cannot restrict memory")

            if self.timeout:
                # time may be restricted
                resource.setrlimit(resource.RLIMIT_CPU, (self.timeout, chard))
            else:
                warnings.warn("Cannot restrict cpu time")

    def _restricted_function(self, result_bucket, args, kwargs):
        try:
            self._restrict()
            result = self.function(*args, **kwargs)
            result_bucket["result"] = result
        except Exception as e:
            msg = "{}\n\nOriginal {}".format(e, traceback.format_exc())
            result_bucket["result"] = e.__class__(msg)

    def run_restricted(self, *args, **kwargs):
        """
        Executes a given function with restricted amount of
        CPU time and RAM memory usage
        """
        manager = multiprocessing.Manager()
        result_bucket = manager.dict()

        rprocess = multiprocessing.Process(
            target=self._restricted_function, args=[result_bucket, args, kwargs]
        )

        rprocess.start()
        rprocess.join()
        result = result_bucket["result"]

        if isinstance(result, Exception):  # Exception ocurred
            raise result

        return result

    def get_used_memory(self):
        """
        returns the amount of memory being used by the current process
        """
        process = psutil.Process(os.getpid())
        return process.memory_info().rss

    def __call__(self, *args, **kwargs):
        return self.run_restricted(*args, **kwargs)


def alarm_handler(*args):
    raise TimeoutError("process %d got to time limit" % os.getpid())


class RestrictedWorkerByJoin(RestrictedWorker):
    def __init__(self, function, timeout: int, memory: int):
        self.function = function
        self.timeout = timeout
        self.memory = memory

    def _restrict(self):
        if platform.system() == "Linux":
            _, mhard = resource.getrlimit(resource.RLIMIT_AS)
            used_memory = self.get_used_memory()

            if self.memory is None:
                return

            if self.memory > (used_memory + 50 * Mb):
                # memory may be restricted
                logger.info("💻 Restricting memory to %s" % self.memory)
                resource.setrlimit(resource.RLIMIT_DATA, (self.memory, mhard))
            else:
                raise ValueError(
                    "Cannot restrict memory to %s < %i"
                    % (self.memory, used_memory + 50 * Mb)
                )

    def run_restricted(self, *args, **kwargs):
        """
        Executes a given function with restricted amount of
        CPU time and RAM memory usage
        """
        manager = multiprocessing.Manager()
        result_bucket = manager.dict()

        rprocess = multiprocessing.Process(
            target=self._restricted_function, args=[result_bucket, args, kwargs]
        )

        rprocess.start()
        rprocess.join(self.timeout)

        if rprocess.exitcode == 0:
            result = result_bucket["result"]
        else:
            rprocess.terminate()
            raise TimeoutError()

        if isinstance(result, Exception):  # Exception ocurred
            raise result

        return result



from autogoal.kb import *
from autogoal.grammar import CategoricalValue, BooleanValue
from autogoal.utils import nice_repr
from autogoal.kb import AlgorithmBase

import numpy as np


@nice_repr
class VectorAggregator(AlgorithmBase):
    def __init__(self, mode: CategoricalValue("mean", "max")):
        self.mode = mode

    def run(self, input: Seq[VectorContinuous]) -> VectorContinuous:
        input = np.vstack(input)

        if self.mode == "mean":
            return input.mean(axis=1)
        elif self.mode == "max":
            return input.max(axis=1)

        raise ValueError("Invalid mode: %s" % self.mode)


@nice_repr
class MatrixBuilder(AlgorithmBase):
    """
    Builds a matrix from a list of vectors.

    ##### Examples

    ```python
    >>> import numpy as np
    >>> x1 = np.asarray([1,2,3])
    >>> x2 = np.asarray([2,3,4])
    >>> x3 = np.asarray([3,4,5])
    >>> MatrixBuilder().run([x1, x2, x3])
    array([[1, 2, 3],
           [2, 3, 4],
           [3, 4, 5]])

    ```
    """

    def run(self, input: Seq[VectorContinuous]) -> MatrixContinuousDense:
        return np.vstack(input)


@nice_repr
class TensorBuilder(AlgorithmBase):
    """
    Builds a 3D tensor from a list of matrices.

    ##### Examples

    ```python
    >>> import numpy as np
    >>> x1 = np.asarray([[1,2],[3,4]])
    >>> x2 = np.asarray([[2,3],[4,5]])
    >>> x3 = np.asarray([[3,4],[5,6]])
    >>> TensorBuilder().run([x1, x2, x3])
    array([[[1, 2],
            [3, 4]],
    <BLANKLINE>
           [[2, 3],
            [4, 5]],
    <BLANKLINE>
           [[3, 4],
            [5, 6]]])

    ```
    """

    def run(self, input: Seq[MatrixContinuousDense]) -> Tensor3:
        return np.vstack([np.expand_dims(m, axis=0) for m in input])


@nice_repr
class FlagsMerger(AlgorithmBase):
    def run(self, input: Seq[FeatureSet]) -> FeatureSet:
        result = {}

        for d in input:
            result.update(d)

        return result


@nice_repr
class MultipleFeatureExtractor(AlgorithmBase):
    def __init__(
        self,
        extractors: Distinct(
            algorithm(Word, FeatureSet), exceptions=["MultipleFeatureExtractor"]
        ),
        merger: algorithm(Seq[FeatureSet], FeatureSet),
    ):
        self.extractors = extractors
        self.merger = merger

    def run(self, input: Word) -> FeatureSet:
        flags = [extractor.run(input) for extractor in self.extractors]
        return self.merger.run(flags)


@nice_repr
class SentenceFeatureExtractor(AlgorithmBase):
    def __init__(
        self,
        tokenizer: algorithm(Sentence, Seq[Word]),
        feature_extractor: algorithm(Word, FeatureSet),
        include_text: BooleanValue(),
    ):
        self.tokenizer = tokenizer
        self.feature_extractor = feature_extractor
        self.include_text = include_text

    def run(self, input: Sentence) -> FeatureSet:
        tokens = self.tokenizer.run(input)
        flags = [self.feature_extractor(w) for w in tokens]

        if self.include_text:
            return {
                f"{w}|{f}": v for w, flag in zip(tokens, flags) for f, v in flag.items()
            }
        else:
            return {f: v for flag in flags for f, v in flag.items()}


@nice_repr
class DocumentFeatureExtractor(AlgorithmBase):
    def __init__(
        self,
        tokenizer: algorithm(Document, Seq[Sentence]),
        feature_extractor: algorithm(Sentence, FeatureSet),
    ):
        self.tokenizer = tokenizer
        self.feature_extractor = feature_extractor

    def run(self, input: Document) -> Seq[FeatureSet]:
        tokens = self.tokenizer.run(input)
        flags = [self.feature_extractor(w) for w in tokens]
        return


# @nice_repr
# class TextEntityEncoder(AlgorithmBase):
#     """
#     Convierte una oración en texto plano más la lista de entidades
#     reconocidas en la oración, en una lista de tokens con sus respectivas
#     categorias BILOUV.
#     """

#     def __init__(self, tokenizer: algorithm(Sentence, Seq[Word])):
#         self.tokenizer = tokenizer

#     def run(
#         self, input: Tuple(Sentence, Seq[Entity])
#     ) -> Tuple(Seq[Word], Seq[Postag]:
#         pass


# @nice_repr
# class TextRelationEncoder(AlgorithmBase):
#     """
#     Convierte una oración en texto plano y una lista de relaciones
#     que se cumplen entre entidades, en una lista de ejemplos
#     por cada oración.
#     """

#     def __init__(self,
#         tokenizer: algorithm(Sentence, Seq[Word]),
#         token_feature_extractor: algorithm(Word, FeatureSet),
#         # token_sentence_encoder: algorithm(Word, )
#     ):
#         pass

#     def run(
#         self, input: Tuple(Sentence, Seq[Tupl](Entity(), Entity(), Category())))
#     ) -> Tuple(Seq[Vect]r()), CategoricalVector()):
#         pass



from collections import namedtuple
import inspect
import abc
import types
import warnings
from typing import Any, Dict, List, Tuple, Type
import types

import networkx as nx
from autogoal.utils import nice_repr
from autogoal.grammar import Graph, GraphSpace, generate_cfg
from autogoal.kb._semantics import SemanticType, Seq


class Supervised(SemanticType):
    """Represents a supervised version of some type X.
    
    It is considered a subclass of X for semantic purposes, but not the other way around:

    # >>> issubclass(Supervised[Vector], Vector)
    # True
    # >>> issubclass(Vector, Supervised[Vector])
    # False
    # >>> issubclass(Supervised[Seq[Vector]], Seq[Vector])
    # True
    # >>> issubclass(Seq[Vector], Supervised[Seq[Vector]])
    # False
    
    """

    __internal_types = {}

    @classmethod
    def _specialize(cls, internal_type):
        try:
            return cls.__internal_types[internal_type]
        except KeyError:
            pass

        class SupervisedImp(Supervised):
            __internal = internal_type

            @classmethod
            def _name(cls):
                return f"Supervised[{cls.__internal}]"

            @classmethod
            def _reduce(cls):
                return Supervised._specialize, (internal_type,)

        cls.__internal_types[internal_type] = SupervisedImp

        return SupervisedImp


def algorithm(*annotations):
    from autogoal.grammar import Union, Symbol

    *inputs, output = annotations

    def match(cls):
        if not hasattr(cls, "run"):
            return False

        signature = inspect.signature(cls.run)
        input_types = [
            v.annotation for k, v in signature.parameters.items() if k != "self"
        ]
        output_type = signature.return_annotation

        if len(inputs) != len(input_types):
            return False

        for expected, real in zip(inputs, input_types):
            if not issubclass(expected, real):
                return False

        if not issubclass(output_type, output):
            return False

        return True

    @classmethod
    def generate_cfg(cls, grammar, head):
        symbol = head or Symbol(cls.__name__)
        compatible = []

        for _, other_cls in grammar.namespace.items():
            if match(other_cls):
                compatible.append(other_cls)

        if not compatible:
            raise ValueError(
                f"Cannot find any suitable implementation of algorithms with inputs: {inputs} and output: {output}"
            )

        return Union(symbol.name, *compatible).generate_cfg(grammar, symbol)

    def build(ns):
        ns["generate_cfg"] = generate_cfg

    return types.new_class(f"Algorithm[{inputs},{output}]", bases=(), exec_body=build)


class Algorithm(abc.ABC):
    """Represents an abstract algorithm with a run method.

    Provides introspection for the expected semantic input and output types.
    Users should inherit from `AlgorithmBase` instead of this class.
    """

    @abc.abstractclassmethod
    def input_types(cls) -> Tuple[type]:
        """Returns an ordered list of the expected semantic input types of the `run` method.
        """
        pass

    @abc.abstractclassmethod
    def input_args(cls) -> Tuple[str]:
        """Returns an ordered tuple of the names of the arguments in the `run` method.
        """
        pass

    @abc.abstractclassmethod
    def output_type(cls) -> type:
        """Returns an ordered list of the expected semantic output type of the `run` method.
        """
        pass

    @abc.abstractmethod
    def run(self, *args):
        """Executes the algorithm.
        """
        pass

    @classmethod
    def is_compatible_with(cls: "Algorithm", input_types):
        """
        Determines if the current algorithm is compatible with a set of input types,
        i.e., if among those types we can find all the necessary inputs for this algorithm.

        >>> class A(AlgorithmBase):
        ...     def run(self, x:int) -> float:
        ...         pass
        >>> A.is_compatible_with([int])
        True
        """
        my_inputs = cls.input_types()

        for needed in my_inputs:
            for having in input_types:
                if issubclass(having, needed):
                    break
            else:
                return False

        return True


class AlgorithmBase(Algorithm):
    """Represents an algorithm,

    Automatically implements the input and output introspection methods using the `inspect` module.
    Users inheriting from this class must provide type annotations in the `run` method.
    """

    @classmethod
    def input_types(cls) -> Tuple[type]:
        if not hasattr(cls, "__run_signature__"):
            cls.__run_signature__ = inspect.signature(cls.run)

        return tuple(
            param.annotation
            for name, param in cls.__run_signature__.parameters.items()
            if name != "self"
        )

    @classmethod
    def input_args(cls) -> Tuple[str]:
        if not hasattr(cls, "__run_signature__"):
            cls.__run_signature__ = inspect.signature(cls.run)

        return tuple(
            name for name in cls.__run_signature__.parameters if name != "self"
        )

    @classmethod
    def output_type(cls) -> type:
        if not hasattr(cls, "__run_signature__"):
            cls.__run_signature__ = inspect.signature(cls.run)

        return cls.__run_signature__.return_annotation


def build_input_args(algorithm: Algorithm, values: Dict[type, Any]):
    """Buils the correct input mapping for `algorithm` using the provided `values` mapping types to objects.

    The input can be a class that inherits from `Algorithm` or an instance of such a class.

    >>> class A(AlgorithmBase):
    ...    def run(self, a:int, b:str):
    ...        pass
    >>> values = { str:"hello", float:3.0, int:42 }
    >>> build_input_args(A, values)
    {'a': 42, 'b': 'hello'}
    >>> build_input_args(A(), values)
    {'a': 42, 'b': 'hello'}

    """

    result = {}

    for name, type in zip(algorithm.input_args(), algorithm.input_types()):
        try:
            result[name] = values[type]
        except KeyError:
            for key in values:
                if issubclass(key, type):
                    result[name] = values[key]
                    break
            else:
                raise TypeError(f"Cannot find compatible input value for {type}")

    return result


@nice_repr
class Pipeline:
    """Represents a sequence of algorithms.

    Each algorithm must have a `run` method declaring it's input and output type.
    The pipeline instance also receives the input and output types.
    """

    def __init__(
        self, algorithms: List[Algorithm], input_types: List[Type[SemanticType]]
    ) -> None:
        self.algorithms = algorithms
        self.input_types = input_types

    def run(self, *inputs):
        data = {}

        for i, t in zip(inputs, self.input_types):
            data[t] = i

        for algorithm in self.algorithms:
            args = build_input_args(algorithm, data)
            output = algorithm.run(**args)
            output_type = algorithm.output_type()
            data[output_type] = output

        return data[self.algorithms[-1].output_type()]

    def send(self, msg: str, *args, **kwargs):
        found = False

        for step in self.algorithms:
            if hasattr(step, msg):
                getattr(step, msg)(*args, **kwargs)
                found = True
            elif hasattr(step, "send"):
                step.send(msg, *args, **kwargs)
                found = True

        if not found:
            warnings.warn(f"No step answered message {msg}.")


def make_seq_algorithm(algorithm: Algorithm):
    """Lift an algorithm with input types T1, T2, Tn to a meta-algorithm with types Seq[T1], Seq[T2], ...

    The generated class correctly defines the input and output types.
    These implementations are compatible with `build_input_args`:

    >>> class A(AlgorithmBase):
    ...     def __init__(self, alpha):
    ...         self.alpha = 0.5
    ...     def run(self, x:int, y:str) -> float:
    ...         return self.alpha * (x + len(y))
    ...     def __repr__(self):
    ...         return f"A({self.alpha})"
    >>> B = make_seq_algorithm(A)
    >>> b = B(0.5)
    >>> b
    SeqAlgorithm[A(0.5)]
    >>> b.run([1, 2], y=["A", "BC"])
    [1.0, 2.0]
    >>> B.input_types()
    (Seq[<class 'int'>], Seq[<class 'str'>])
    >>> B.input_args()
    ('x', 'y')
    >>> b.output_type()
    Seq[<class 'float'>]
    >>> build_input_args(B, {Seq[int]: [1, 2], Seq[str]: ["hello", "world"]})
    {'x': [1, 2], 'y': ['hello', 'world']}
    
    """

    output_type = algorithm.output_type()

    name = f"SeqAlgorithm[{algorithm.__name__}]"

    def init_method(self, *args, **kwargs):
        self.inner = algorithm(*args, **kwargs)

    def run_method(self, *args, **kwargs) -> Seq[output_type]:
        args_kwargs = _make_list_args_and_kwargs(*args, **kwargs)
        return [self.inner.run(*t.args, **t.kwargs) for t in args_kwargs]

    def repr_method(self):
        return f"SeqAlgorithm[{repr(self.inner)}]"

    def getattr_method(self, attr):
        return getattr(self.inner, attr)

    @classmethod
    def input_types_method(cls):
        return tuple(Seq[t] for t in algorithm.input_types())

    @classmethod
    def input_args_method(cls):
        return algorithm.input_args()

    @classmethod
    def output_types_method(cls):
        return Seq[algorithm.output_type()]

    def body(ns):
        ns["__init__"] = init_method
        ns["run"] = run_method
        ns["__repr__"] = repr_method
        ns["__getattr__"] = getattr_method
        ns["input_types"] = input_types_method
        ns["input_args"] = input_args_method
        ns["output_type"] = output_types_method

    return types.new_class(name=name, bases=(Algorithm,), exec_body=body)


Akw = namedtuple("Akw", ["args", "kwargs"])


def _make_list_args_and_kwargs(*args, **kwargs):
    """Transforms a list of args into individual args and kwargs for an internal algorithm. 
    
    To be used by `make_seq_algorithm"

    >>> _make_list_args_and_kwargs([1,2], [4,5])
    [Akw(args=(1, 4), kwargs={}), Akw(args=(2, 5), kwargs={})]
    >>> _make_list_args_and_kwargs(x=[1,2], y=[4,5])
    [Akw(args=(), kwargs={'x': 1, 'y': 4}), Akw(args=(), kwargs={'x': 2, 'y': 5})]
    >>> _make_list_args_and_kwargs([1,2], y=[4,5])
    [Akw(args=(1,), kwargs={'y': 4}), Akw(args=(2,), kwargs={'y': 5})]

    """
    lengths = set(len(v) for v in kwargs.values()) | set(len(v) for v in args)

    if len(lengths) != 1:
        raise ValueError("All args and kwargs must be sequences of the same length.")

    length = lengths.pop()

    inner_args = []

    for i in range(length):
        inner_args.append(tuple([xs[i] for xs in args]))

    inner_kwargs = []

    for i in range(length):
        inner_kwargs.append({k: v[i] for k, v in kwargs.items()})

    return [Akw(xs, ks) for xs, ks in zip(inner_args, inner_kwargs)]


class PipelineNode:
    def __init__(self, algorithm, input_types, output_types, registry=None) -> None:
        self.algorithm = algorithm
        self.input_types = set(input_types)
        self.output_types = set(output_types)
        self.grammar = generate_cfg(self.algorithm, registry=registry)

    def sample(self, sampler):
        return self.grammar.sample(sampler=sampler)

    @property
    def __name__(self):
        return self.algorithm.__name__

    def __eq__(self, o: object) -> bool:
        return isinstance(o, PipelineNode) and all(
            [o.algorithm == self.algorithm, o.input_types == self.input_types,]
        )

    def __repr__(self) -> str:
        return f"<PipelineNode(algorithm={self.algorithm.__name__},input_types={[i.__name__ for i in self.input_types]},output_types={[o.__name__ for o in self.output_types]})>"

    def __hash__(self) -> int:
        return hash(repr(self))


class PipelineSpace(GraphSpace):
    def __init__(self, graph: Graph, input_types):
        super().__init__(graph, initializer=self.initialize)
        self.input_types = input_types

    def initialize(self, item: PipelineNode, sampler):
        return item.sample(sampler)

    def sample(self, *args, **kwargs):
        path = super().sample(*args, **kwargs)
        return Pipeline(path, input_types=self.input_types)


def build_pipeline_graph(
    input_types: List[type],
    output_type: type,
    registry: List[type],
    max_list_depth: int = 3,
):
    """Build a graph of algorithms.

    Every node in the graph corresponds to a <autogoal.grammar.ContextFreeGrammar> that
    generates an instance of a class with a `run` method.

    Each `run` method must declare input and output types in the form:

        def run(self, a: type_1, b: type_2, ...) -> type_n:
            # ...
    """

    if not isinstance(input_types, (list, tuple)):
        input_types = [input_types]

    # We start by enlarging the registry with all Seq[...] algorithms

    pool = set(registry)

    for algorithm in registry:
        for _ in range(max_list_depth):
            algorithm = make_seq_algorithm(algorithm)
            pool.add(algorithm)

    # For building the graph, we'll keep at each node the guaranteed output types

    # We start by collecting all the possible input nodes,
    # those that can process a subset of the input_types
    open_nodes: List[PipelineNode] = []

    for algorithm in pool:
        if not algorithm.is_compatible_with(input_types):
            continue

        open_nodes.append(
            PipelineNode(
                algorithm=algorithm,
                input_types=input_types,
                output_types=set(input_types) | set([algorithm.output_type()]),
                registry=registry,
            )
        )

    G = Graph()

    for node in open_nodes:
        G.add_edge(GraphSpace.Start, node)

    # We'll make a BFS exploration of the pipeline space.
    # For every open node we will add to the graph every node to which it can connect.
    closed_nodes = set()

    while open_nodes:
        node = open_nodes.pop(0)

        # These are the types that are available at this node
        guaranteed_types = node.output_types

        # Here are all the algorithms that could be added new at this point in the graph
        for algorithm in registry:
            if not algorithm.is_compatible_with(guaranteed_types):
                continue

            # We never want to apply the same exact algorithm twice
            if algorithm == node.algorithm:
                continue

            # And we never want an algorithm that doesn't provide a novel output type...
            if (
                algorithm.output_type() in guaranteed_types
                and
                # ... unless it is an idempotent algorithm
                [algorithm.output_type()] != algorithm.input_types()
            ):
                continue

            p = PipelineNode(
                algorithm=algorithm,
                input_types=guaranteed_types,
                output_types=guaranteed_types | set([algorithm.output_type()]),
                registry=registry,
            )

            G.add_edge(node, p)

            if p not in closed_nodes and p not in open_nodes:
                open_nodes.append(p)

        # Now we check to see if this node is a possible output
        if issubclass(node.algorithm.output_type(), output_type):
            G.add_edge(node, GraphSpace.End)

        closed_nodes.add(node)

    # Remove all nodes that are not connected to the end node
    try:
        reachable_from_end = set(
            nx.dfs_preorder_nodes(G.reverse(False), GraphSpace.End)
        )
        unreachable_nodes = set(G.nodes) - reachable_from_end
        G.remove_nodes_from(unreachable_nodes)
    except KeyError:
        raise TypeError("No pipelines can be found!")

    return PipelineSpace(G, input_types=input_types)


__all__ = [
    "AlgorithmBase",
    "Supervised",
    "Pipeline",
    "build_pipeline_graph",
    "algorithm",
]


# Finally, we run doctest in the module for easy testing of the functional API.

if __name__ == "__main__":
    import doctest

    doctest.testmod()



from ._base import (
    SearchAlgorithm,
    ProgressLogger,
    ConsoleLogger,
    Logger,
    MemoryLogger,
    RichLogger,
)
from ._random import RandomSearch
from ._pge import ModelSampler, PESearch
from ._learning import SurrogateSearch



# from autogoal.contrib import ensemble
import numpy as np
from sklearn.svm import SVC
from sklearn.naive_bayes import GaussianNB
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier


class StackingEnsemble:
    def __init__(self, layers=None, final=None):

        if layers == None:
            self.layers = [[SVC(), LogisticRegression()], [DecisionTreeClassifier()]]

        else:
            self.layers = layers

        if final == None:
            self.final = GaussianNB()

        else:
            self.final = final

        self.network = []

    def network_constructor(self):
        """
        Creates a network containing layers of estimators.
        """
        network = self.network
        layers = self.layers
        final = self.final
        network.append(layers)
        network.append(final)
        return network

    def forward_pass(self, X, y):
        """
        Do a forward pass of the stacked network
        """

        network = self.network_constructor()

        output = y
        input_current_layer = []
        input_next_layer = []

        for index, layer in enumerate(network):

            if index == 0:
                input_current_layer = X
                for estimator in layer:
                    estimator.fit(input_current_layer, output)
                    input_next_layer.append(estimator.predict(input_current_layer))
            else:
                input_current_layer = input_next_layer
                input_next_layer = []
                for estimator in layer:
                    estimator.fit(input_current_layer, output)
                    input_next_layer.append(estimator.predict(input_current_layer))

        return network

    def fit(self, X, y):
        input_length = len(X)
        target_lenght = len(y)
        if input_length == target_lenght:
            return self.forward_pass(X, y)
        else:
            raise ValueError("X and y must have the same length")

    def predict(self, X):
        """
        Do a prediction for a test data
        """
        network = self.network
        prediction_current_layer = np.array([])
        input_current_layer = []
        for index, layer in enumerate(network):

            if index == 0:
                input_current_layer = X
                for estimator in layer:
                    prediction_current_layer = np.concatenate(
                        (
                            prediction_current_layer,
                            estimator.predict(input_current_layer),
                        )
                    )

                prediction_current_layer = np.reshape(prediction_current_layer, (1, 2))

            else:
                input_current_layer = prediction_current_layer
                prediction_current_layer = np.array([])
                for estimator in layer:
                    prediction_current_layer = np.concatenate(
                        (
                            prediction_current_layer,
                            estimator.predict(input_current_layer),
                        )
                    )

        return prediction_current_layer


if __name__ == "__main__":
    X_train = [[0, 0], [1, 1]]
    y_train = [0, 1]
    X_test = [[2.0, 2.0]]
    y_test = [1]
    ensemble = StackingEnsemble([SVC(), DecisionTreeClassifier()], [SVC()])
    ensemble.fit(X_train, y_train)
    y_pred = ensemble.predict(X_test)



import os
from pathlib import Path

CONTRIB_NAME = "gensim"
DATA_PATH = Path.home() / ".autogoal" / "contrib" / CONTRIB_NAME / "data"

# ensure data path directory creation
try:
    os.makedirs(DATA_PATH)
except IOError as ex:
    # directory already exists
    pass


def load_data(path, name):
    use_data(path)
    try:
        import gensim
        import gensim.downloader as api

        api.load(name)
    except:
        pass


def use_data(path):
    os.environ["GENSIM_DATA_DIR"] = str(path)



import random

from autogoal.datasets import download, datapath


def load(max_examples=None):
    try:
        download("movie_reviews")
    except:
        print(
            "Error loading data. This may be caused due to bad connection. Please delete badly downloaded data and retry"
        )
        raise

    sentences = []
    classes = []

    path = datapath("movie_reviews")

    ids = list(path.rglob("*.txt"))
    random.shuffle(ids)

    for fd in ids:
        if "neg/" in str(fd):
            cls = "neg"
        else:
            cls = "pos"

        with fd.open() as fp:
            sentences.append(fp.read())
            classes.append(cls)

        if max_examples and len(classes) >= max_examples:
            break

    return sentences, classes


def make_fn(test_size=0.25, examples=None):
    from sklearn.model_selection import train_test_split
    from sklearn.metrics import accuracy_score

    X, y = load(examples)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size)

    def fitness_fn(pipeline):
        pipeline.fit(X_train, y_train)
        y_pred = pipeline.predict(X_test)

        return accuracy_score(y_test, y_pred)

    return fitness_fn



"""Defines the semantic types hierarchy and utilities to infer types and work with them.
"""

# The AutoML features in AutoGOAL rest in the ability of automatically discovering pipelines
# of algorithms that are semantically annotated w.r.t. their inputs and outputs.
# This module defines the semantic types hierarchy.
#
# A semantic type is basically a class that provides a meaningful
# definition for a value in the problem domain. For example, a list of strings can either
# be a list of sentences, words, DNA sequences, or full-fledged documents. Depending on the
# interpretation of that value, we want AutoGOAL to select different algorithms, i.e., it makes
# no sense to tokenize words, while it is almost mandatory to do so when you have natural language text.

# We will never really instantiate these classes, just use them for annotations.

from functools import reduce
import inspect
import copyreg
from typing import Type


# We start by defining the base class of our hierarchy.
# A `SemanticType` is just a class that knows how to do one thing: determine if a given object
# matches its semantic definition.
# We will define a metaclass to allow for `isinstance` and `issubclass` to work based on our semantic match definition,
# and to leave space for implementing class-level `__getitem__` to allow for a generics kind-of notation.


class SemanticTypeMeta(type):
    def __instancecheck__(cls, instance) -> bool:
        return cls._match(instance)

    def __getitem__(cls, args):
        if isinstance(args, tuple):
            return cls._specialize(*args)
        else:
            return cls._specialize(args)

    def __subclasscheck__(cls, subclass: type) -> bool:
        if hasattr(subclass, "_conforms") and subclass._conforms(cls):
            return True

        return super().__subclasscheck__(subclass)

    def __call__(self, *args, **kwds):
        raise TypeError("Cannot instantiate a semantic type")

    def __repr__(cls) -> str:
        return cls._name()


# `SemanticType` defines a `match` method that implements our `isinstance` method.


class SemanticType(metaclass=SemanticTypeMeta):
    @classmethod
    def _match(cls, x) -> bool:
        return False

    @classmethod
    def _conforms(cls, other: type) -> bool:
        return False

    @classmethod
    def _name(cls) -> str:
        return cls.__name__

    @classmethod
    def _specialize(cls, *args):
        raise TypeError(f"{cls} cannot be specialized")

    @classmethod
    def _reduce(cls):
        # By default, we return the repr name of the class that
        # allows to serialize globally defined classes.
        return cls._name()

    @staticmethod
    def infer(x):
        """Automatically determines the semantic type of a given value.
        
        >>> SemanticType.infer("word")
        Word
        >>> SemanticType.infer("hello world")
        Sentence
        >>> SemanticType.infer("Hello world. This a two sentence Document.")
        Document

        It works with some tensorial types as well.

        >>> import numpy as np
        >>> SemanticType.infer(np.ones(shape=(2,2)))
        Tensor[2, Continuous, Dense]

        """
        types = inspect.getmembers(inspect.getmodule(SemanticType), inspect.isclass)
        best_type = SemanticType

        for _, t in types:
            if isinstance(x, t) and issubclass(t, best_type):
                best_type = t

        if best_type == SemanticType:
            raise ValueError(f"Cannot infer semantic type for {x}")

        return best_type


# To be able to serialize these types, we have to register a reduce function for `SemanticTypeMeta`.
# This reduce function will just dispatch to the proper instance method


def _reduce_semantic_type(t):
    return t._reduce()


copyreg.pickle(SemanticTypeMeta, _reduce_semantic_type)


# Let's start with the natural language hierarchy.


class Text(SemanticType):
    @classmethod
    def _match(cls, x):
        return isinstance(x, str)


class Document(Text):
    pass


class Sentence(Document):
    @classmethod
    def _match(cls, x):
        return super()._match(x) and x.count(".") <= 1


class Word(Sentence):
    @classmethod
    def _match(cls, x):
        return super()._match(x) and " " not in x


# We also need some basic types for generic kinds of labels, used in NLP, for example.
# Keep in mind these do not implement `_match` as there is no sensible structural way to define them.
# They are used solely to annotate algorithms.


class Label(SemanticType):
    pass


class Postag(Label):
    pass


class Chunktag(Label):
    pass


class Stem(Text):
    pass


class Synset(SemanticType):
    pass


class Sentiment(Label):
    pass


# Another basic type is a feature set, which for us is basically a dictionary of strings vs anything:


class FeatureSet(SemanticType):
    @classmethod
    def _match(self, x):
        # TODO: This is a very naive implementation of `match`.
        return isinstance(x, dict) and isinstance(list(x.keys()[0]), str)


# A first complex type we can implement is `Seq`, to represent a list (or sequence) of another semantic type.
# We want this type to be able to specialize in this notation:
#
# >>> Seq[Word]
#
# For this we have to implement `_specialize` and synthethize a new type with the corresponding
# semantic type that does the match internally.
# The challenge here is that we want to return the same `Seq` type every time we specialize on the same internal type,
# so we'll keep a class-level dictionary to store these classes as they are synthethized.
#
# Finally, we want `Seq[Word]`` to be a subclass of `Seq[Sentence]`, or more generally, `Seq[X]` to be a subclass
# of `Seq[Y]` whenever `X < Y`.


class Seq(SemanticType):
    """Represents a sequence that can be specialized in concrete internal semantic types.

    >>> isinstance(["hello", "world"], Seq[Word])
    True

    Specialized classes are exactly the same, by identity:

    >>> id(Seq[Word]) == id(Seq[Word])
    True

    Specialized classes are subclasses of the raw `Seq` class:

    >>> issubclass(Seq[Word], Seq)
    True

    And they are subclasses of other specialized classes when the internal types hold the same relationship.

    >>> issubclass(Seq[Word], Seq[Text])
    True
    >>> issubclass(Seq[Text], Seq[Word])
    False

    They are pretty-printed as well

    >>> Seq
    Seq
    >>> Seq[Word]
    Seq[Word]

    Subclasses are also serializable (which requires some non-trivial dark magic on the implementation side):

    >>> from pickle import dumps, loads
    >>> loads(dumps(Seq[Word]))
    Seq[Word]
   
    """

    __internal_types = {}

    @classmethod
    def _specialize(cls, internal_type: Type[SemanticType]):
        try:
            return Seq.__internal_types[internal_type]
        except KeyError:
            pass

        class SeqImp(Seq):
            __internal_type__ = internal_type

            @classmethod
            def _name(cls):
                return f"Seq[{internal_type}]"

            @classmethod
            def _match(cls, x):
                return isinstance(x, (list, tuple)) and internal_type._match(x[0])

            @classmethod
            def _conforms(cls, other):
                if not issubclass(other, Seq):
                    return False

                if other == Seq:
                    return True

                return issubclass(cls.__internal_type__, other.__internal_type__)

            @classmethod
            def _specialize(cls, *args, **kwargs):
                raise TypeError("Cannot specialize more a `Seq` type.")

            @classmethod
            def _reduce(cls):
                return Seq._specialize, (internal_type,)

        Seq.__internal_types[internal_type] = SeqImp

        return SeqImp


# Now let's move to the algebraic types, vectors, matrices, and tensors.
# These wrap numpy arrays of different dimensionalities
# We'll have three different semantic labels for each tensor: dimensionality (an integer),
# internal type, and a dense/sparse flag.


from numpy import ndarray
from scipy.sparse.base import spmatrix


# These instances represent the two types of tensorial structure.


class TensorStructure:
    def __init__(self, base_class: type, name: str) -> None:
        self.base_class = base_class
        self.name = name

    def __repr__(self) -> str:
        return self.name

    def match(self, x):
        return isinstance(x, self.base_class)

    def __eq__(self, o: object) -> bool:
        return isinstance(o, TensorStructure) and o.name == self.name

    def __hash__(self) -> int:
        return hash(self.name)


Dense = TensorStructure(ndarray, "Dense")
Sparse = TensorStructure(spmatrix, "Sparse")


class TensorData:
    def __init__(self, dtype_label: str, name: str) -> None:
        self.dtype_label = dtype_label
        self.name = name

    def __repr__(self) -> str:
        return self.name

    def match(self, x):
        return x.dtype.kind == self.dtype_label

    def __eq__(self, o: object) -> bool:
        return isinstance(o, TensorData) and o.name == self.name

    def __hash__(self) -> int:
        return hash(self.name)


Categorical = TensorData("U", "Categorical")
Continuous = TensorData("f", "Continuous")
Discrete = TensorData("i", "Discrete")


# We want the abstract `Tensor` type to be specializable using the notation:
# `Tensor[3, Category, Dense]`.
# This requires us to implement the `_specialize` just like we did with `Seq`.
# The internal class will in turn implement `_match` accordingly to how those values are defined.

# One special thing we want to do is to let some of these semantic flags undefined (using `None`)
# such that `Tensor[2, None, Dense]` represents any structure that can have two dimensions no matter the internal type.
# And we want `issubclass(...)` to work in a way that `Tensor[2, Categorical, Dense]` is a subclass to `Tensor[2, None, Dense]`.
# For this purpose we will redefine `_conforms` to match according to how those semantic flags are defined.


class Tensor(SemanticType):
    """Represents an abstract tensor type. Can be specialized into more concrete types.

    They support type inference from numpy:

    >>> import numpy as np
    >>> a_matrix = np.ones(shape=(2,2))
    >>> isinstance(a_matrix, Tensor)
    True
    >>> isinstance(a_matrix, Tensor[1, None, None])
    False
    >>> isinstance(a_matrix, Tensor[2, None, None])
    True
    >>> isinstance(a_matrix, Tensor[2, Continuous, Dense])
    True
    >>> isinstance(a_matrix, Tensor[2, Continuous, Sparse])
    False

    Tensor types also respect a special definition of subclass in which more specialized
    classes are defined as subclasses of less specialized counterparts.

    >>> issubclass(Tensor[2, Continuous, Sparse], Tensor[2, None, None])
    True
    >>> issubclass(Tensor[2, Continuous, Sparse], Tensor[2, Continuous, None])
    True
    >>> issubclass(Tensor[2, Continuous, Sparse], Tensor[2, None, Dense])
    False

    Each specific specialization is a singleton class:
    
    >>> id(Tensor[2, Continuous, Dense]) == id(Tensor[2, Continuous, Dense])
    True

    Tensor types are also serializable:

    >>> from pickle import loads, dumps
    >>> loads(dumps(Tensor[2, Continuous, Dense]))
    Tensor[2, Continuous, Dense]

    """

    __internal_types = {}

    @classmethod
    def _match(self, x):
        return isinstance(x, (ndarray, spmatrix))

    @classmethod
    def _specialize(
        cls, dimension: int, internal_type: TensorData, structure: TensorStructure
    ):
        try:
            return Tensor.__internal_types[(dimension, internal_type, structure)]
        except KeyError:
            pass

        class TensorImp(Tensor):
            __flags = (dimension, internal_type, structure)

            @classmethod
            def _name(cls):
                return f"Tensor[{cls.__flags[0]}, {cls.__flags[1]}, {cls.__flags[2]}]"

            @classmethod
            def _match(cls, x):
                if not super()._match(x):
                    return False

                if dimension is not None and (len(x.shape) != dimension):
                    return False

                if internal_type is not None and not internal_type.match(x):
                    return False

                if structure is not None and not structure.match(x):
                    return False

                return True

            @classmethod
            def _conforms(cls, other: type) -> bool:
                if other == Tensor:
                    return True

                if not hasattr(other, "_TensorImp__flags"):
                    return False

                for fmine, fother in zip(cls.__flags, other.__flags):
                    if fother is None:
                        continue

                    if fmine is None:
                        return False

                    if fmine != fother:
                        return False

                return True

            @classmethod
            def _reduce(cls):
                # To reduce Tensor implementations for pickling
                return Tensor._specialize, (dimension, internal_type, structure)

        Tensor.__internal_types[(dimension, internal_type, structure)] = TensorImp

        return TensorImp


# Now that we have the basic tensorial type implemented, we can add some aliases here.
# These aliases mostly serve for `SemanticType.infer` to work, and also to simplify imports,
# but keep in mind that anywhere we use `Vector` we just as well use `Tensor[1, None, None]`,
# as they are *exactly* the same class.

Vector = Tensor[1, None, None]
VectorContinuous = Tensor[1, Continuous, None]
VectorCategorical = Tensor[1, Categorical, Dense]
VectorDiscrete = Tensor[1, Discrete, Dense]

Matrix = Tensor[2, None, None]
MatrixContinuous = Tensor[2, Continuous, None]
MatrixContinuousDense = Tensor[2, Continuous, Dense]
MatrixContinuousSparse = Tensor[2, Continuous, Sparse]
MatrixCategorical = Tensor[2, Categorical, Dense]
MatrixDiscrete = Tensor[2, Discrete, Dense]

# 📝 Makes no sense to have sparse discrete or categorical tensors as you cannot have missing categories.

Tensor3 = Tensor[3, Continuous, Dense]
Tensor4 = Tensor[4, Continuous, Dense]

# Finally we define the publicly export classes

__all__ = [
    "SemanticType",
    "Seq",
    "Text",
    "Document",
    "Sentence",
    "Word",
    "Label",
    "Postag",
    "Chunktag",
    "Stem",
    "Synset",
    "Sentiment",
    "FeatureSet",
    "Tensor",
    "Vector",
    "VectorContinuous",
    "VectorCategorical",
    "VectorDiscrete",
    "Matrix",
    "MatrixContinuous",
    "MatrixContinuousDense",
    "MatrixContinuousSparse",
    "MatrixCategorical",
    "MatrixDiscrete",
    "Tensor3",
    "Tensor4",
    "Dense",
    "Sparse",
    "Categorical",
    "Continuous",
    "Discrete",
]


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from autogoal.exceptions import InterfaceIncompatibleError

# import types
import inspect

# import pprint

from typing import Mapping
from autogoal.grammar import Symbol, Union, Empty, Subset

# from scipy.sparse.base import spmatrix
# from numpy import ndarray


# def algorithm(input_type, output_type):
#     def run_method(self, input: input_type) -> output_type:
#         pass

#     def body(ns):
#         ns["run"] = run_method

#     return types.new_class(
#         name="Algorithm[%s, %s]" % (input_type, output_type),
#         bases=(Interface,),
#         exec_body=body,
#     )


class Interface:
    @classmethod
    def is_compatible(cls, other_cls):
        own_methods = _get_annotations(cls, ignore=["generate_cfg", "is_compatible"])

        if not inspect.isclass(other_cls):
            return False

        if issubclass(other_cls, Interface):
            return False

        type_methods = _get_annotations(other_cls)
        return _compatible_annotations(own_methods, type_methods)

    @classmethod
    def generate_cfg(cls, grammar, head):
        symbol = head or Symbol(cls.__name__)
        compatible = []

        for _, other_cls in grammar.namespace.items():
            if cls.is_compatible(other_cls):
                compatible.append(other_cls)

        if not compatible:
            raise InterfaceIncompatibleError(cls.__name__)

        return Union(symbol.name, *compatible).generate_cfg(grammar, symbol)


class Distinct:
    def __init__(self, interface: Interface, exceptions=None):
        self.interface = interface
        self.exceptions = exceptions

    def generate_cfg(self, grammar, head):
        symbol = head or Symbol(self.__class__.__name__)
        compatible = []

        for _, other_cls in grammar.namespace.items():
            if other_cls in self.exceptions:
                continue

            if hasattr(other_cls, "__name__") and other_cls.__name__ in self.exceptions:
                continue

            if self.interface.is_compatible(other_cls):
                compatible.append(other_cls)

        if not compatible:
            raise ValueError(
                "Cannot find compatible implementations for <class %s>" % self.interface
            )

        return Subset(symbol.name, *compatible).generate_cfg(grammar, symbol)


# def conforms(type1, type2):
#     if inspect.isclass(type1) and inspect.isclass(type2):
#         return issubclass(type1, type2)

#     if hasattr(type1, "__conforms__") and type1.__conforms__(type2):
#         return True

#     if hasattr(type2, "__rconforms__") and type2.__rconforms__(type1):
#         return True

#     return False


def _compatible_annotations(
    methods_if: Mapping[str, inspect.Signature],
    methods_im: Mapping[str, inspect.Signature],
):
    for name, mif in methods_if.items():
        if not name in methods_im:
            return False

        mim = methods_im[name]

        for name, param_if in mif.parameters.items():
            if not name in mim.parameters:
                return False

            param_im = mim.parameters[name]
            ann_if = param_if.annotation

            if ann_if == inspect.Parameter.empty:
                continue

            ann_im = param_im.annotation

            if not conforms(ann_if, ann_im):
                return False

        return_if = mif.return_annotation

        if return_if == inspect.Parameter.empty:
            continue

        return_im = mim.return_annotation

        if not conforms(return_im, return_if):
            return False

    return True


def _get_annotations(clss, ignore=[]):
    """
    Computes the annotations of all public methods in type `clss`.

    ##### Examples

    ```python
    >>> class A:
    ...     def f(self, input: int) -> float:
    ...         pass
    >>> _get_annotations(A)
    {'f': <Signature (self, input:int) -> float>}

    ```
    """
    methods = inspect.getmembers(
        clss, lambda m: inspect.ismethod(m) or inspect.isfunction(m)
    )
    signatures = {
        name: inspect.signature(method)
        for name, method in methods
        if not name.startswith("_")
    }

    for name in ignore:
        signatures.pop(name, None)

    return signatures


# # def make_list_wrapper(algorithm):
# #     from autogoal.kb._algorithm import _get_annotations

# #     input_type, output_type = _get_annotations(algorithm)
# #     name = f"List[{algorithm.__name__}]"

# #     def wrap_list(types):
# #         if isinstance(types, Tuple):
# #             return Tuple(*(List(t) for t in types.inner))

# #         return List(types)

# #     def init_method(self, inner: algorithm):
# #         self.inner = inner

# #     def run_method(self, input: wrap_list(input_type)) -> wrap_list(output_type):
# #         return [self.inner.run(x) for x in xs]

# #     def repr_method(self):
# #         return f"{name}(inner={repr(self.inner)})"

# #     def getattr_method(self, attr):
# #         return getattr(self.inner, attr)

# #     def body(ns):
# #         ns["__init__"] = init_method
# #         ns["run"] = run_method
# #         ns["__repr__"] = repr_method
# #         ns["__getattr__"] = getattr_method

# #     return types.new_class(name=name, bases=(), exec_body=body)


# # def build_composite_list(input_type, output_type, depth=1):
# #     def wrap(t, d):
# #         if d == 0:
# #             return t

# #         return List(wrap(t, d - 1))

# #     input_wrapper = wrap(input_type, depth)
# #     output_wrapper = wrap(output_type, depth)

# #     # name = "ListAlgorithm"  # % (input_wrapper, output_wrapper)
# #     name = "ListAlgorithm" #[%s, %s]" % (input_wrapper, output_wrapper)

# #     def init_method(self, inner: algorithm(input_type, output_type)):
# #         self.inner = inner

# #     def run_method(self, input: input_wrapper) -> output_wrapper:
# #         def wrap_run(xs, d):
# #             if d == 0:
# #                 return self.inner.run(xs)

# #             return [wrap_run(x, d - 1) for x in xs]

# #         return wrap_run(input, depth)

# #     def repr_method(self):
# #         return f"{name}(inner={repr(self.inner)})"

# #     def getattr_method(self, attr):
# #         return getattr(self.inner, attr)

# #     def reduce_method(self):
# #         return (
# #             build_composite_list_instance,
# #             (input_type, output_type, self.inner)
# #         )

# #     def body(ns):
# #         ns["__init__"] = init_method
# #         ns["run"] = run_method
# #         ns["__repr__"] = repr_method
# #         ns["__getattr__"] = getattr_method
# #         ns["__reduce__"] = reduce_method

# #     return types.new_class(name=name, bases=(), exec_body=body)


# # def build_composite_list_instance(input_type, output_type, inner_algorithm):
# #     """
# #     Build a ListAlgorithm[...] type and instantiate it directly on a given algorithm.
# #     """
# #     return build_composite_list(input_type, output_type)(inner_algorithm)


# # def build_composite_tuple(index, input_type: "Tuple", output_type: "Tuple"):
# #     """
# #     Dynamically generate a class `TupleAlgorithm` that wraps
# #     another algorithm to receive a Tuple but pass only one of the
# #     parameters to the internal algorithm.
# #     """

# #     internal_input = input_type.inner[index]
# #     internal_output = output_type.inner[index]

# #     name = "TupleAlgorithm" #[%s, %s]" % (input_type, output_type)

# #     def init_method(self, inner: algorithm(internal_input, internal_output)):
# #         self.inner = inner

# #     def run_method(self, input: input_type) -> output_type:
# #         elements = list(input)
# #         elements[index] = self.inner.run(elements[index])
# #         return tuple(elements)

# #     def repr_method(self):
# #         return f"{name}(inner={repr(self.inner)})"

# #     def getattr_method(self, attr):
# #         return getattr(self.inner, attr)

# #     def reduce_method(self):
# #         return (
# #             build_composite_tuple_instance,
# #             (index, input_type, output_type, self.inner)
# #         )

# #     def body(ns):
# #         ns["__init__"] = init_method
# #         ns["run"] = run_method
# #         ns["__repr__"] = repr_method
# #         ns["__getattr__"] = getattr_method
# #         ns["__reduce__"] = reduce_method

# #     return types.new_class(name=name, bases=(), exec_body=body)


# # def build_composite_tuple_instance(index, input_type, output_type, inner_algorithm):
# #     """
# #     Build a TupleAlgorithm[...] type and instantiate it directly on a given algorithm.
# #     """
# #     return build_composite_tuple(index, input_type, output_type)(inner_algorithm)


# # class DataType:
# #     def __init__(self, **tags):
# #         self.tags = tags

# #     def get_tag(self, tag):
# #         return self.tags.get(tag, None)

# #     def __repr__(self):
# #         # tags = ", ".join(
# #         #     f"{key}={value}"
# #         #     for key, value in sorted(self.tags.items(), key=lambda t: t[0])
# #         # )
# #         return f"{self.__class__.__name__}()" #({tags})"

# #     def __eq__(self, other):
# #         return repr(self) == repr(other)

# #     def __hash__(self):
# #         return hash(repr(self))

# #     @property
# #     def __name__(self):
# #         return self.__class__.__name__

# #     def __conforms__(self, other):
# #         return issubclass(self.__class__, other.__class__)


# # def infer_type(obj):
# #     """
# #     Attempts to automatically infer the most precise semantic type for `obj`.

# #     ##### Parameters

# #     * `obj`: Object to detect its semantic type.

# #     ##### Raises

# #     * `TypeError`: if no valid semantic type was found that matched `obj`.

# #     ##### Examples

# #     * Natural language

# #     ```python
# #     >>> infer_type("hello")
# #     Word()
# #     >>> infer_type("hello world")
# #     Sentence()
# #     >>> infer_type("Hello Word. It is raining.")
# #     Document()

# #     ```

# #     * Vectors

# #     ```
# #     >>> import numpy as np
# #     >>> infer_type(np.asarray(["A", "B", "C", "D"]))
# #     CategoricalVector()
# #     >>> infer_type(np.asarray([0.0, 1.1, 2.1, 0.2]))
# #     ContinuousVector()
# #     >>> infer_type(np.asarray([0, 1, 1, 0]))
# #     DiscreteVector()

# #     ```

# #     * Matrices

# #     ```
# #     >>> import numpy as np
# #     >>> infer_type(np.random.randn(10,10))
# #     MatrixContinuousDense()

# #     >>> import scipy.sparse as sp
# #     >>> infer_type(sp.coo_matrix((10,10)))
# #     MatrixContinuousSparse()

# #     ```
# #     """
# #     if isinstance(obj, str):
# #         if " " not in obj:
# #             return Word()

# #         if "." not in obj:
# #             return Sentence()

# #         return Document()

# #     if isinstance(obj, list):
# #         internal_types = set([infer_type(x) for x in obj])

# #         for test_type in [Document(), Sentence(), Word()]:
# #             if test_type in internal_types:
# #                 return List(test_type)

# #     if hasattr(obj, "shape"):
# #         if len(obj.shape) == 1:
# #             if isinstance(obj, ndarray):
# #                 if obj.dtype.kind == "U":
# #                     return CategoricalVector()
# #                 if obj.dtype.kind == "i":
# #                     return DiscreteVector()
# #                 if obj.dtype.kind == "f":
# #                     return ContinuousVector()
# #         if len(obj.shape) == 2:
# #             if isinstance(obj, spmatrix):
# #                 return MatrixContinuousSparse()
# #             if isinstance(obj, ndarray):
# #                 if obj.dtype.kind == "O":
# #                     return MatrixCategorical()
# #                 else:
# #                     return MatrixContinuousDense()

# #     raise TypeError("Cannot infer type for %r" % obj)


# # class Text(DataType):
# #     pass


# # class Word(Text):
# #     pass


# # class Stem(DataType):
# #     pass


# # class Sentence(Text):
# #     pass


# # class Document(Text):
# #     pass


# # class Category(DataType):
# #     pass


# # class Vector(DataType):
# #     pass


# # class Matrix(DataType):
# #     pass


# # class DenseMatrix(Matrix):
# #     pass


# # class SparseMatrix(Matrix):
# #     pass


# # class ContinuousVector(Vector):
# #     pass


# # class DiscreteVector(Vector):
# #     pass


# # class CategoricalVector(Vector):
# #     pass


# # class MatrixContinuous(Matrix):
# #     pass


# # class MatrixCategorical(Matrix):
# #     pass


# # class MatrixContinuousDense(MatrixContinuous, DenseMatrix):
# #     pass


# # class MatrixContinuousSparse(MatrixContinuous, SparseMatrix):
# #     pass


# # class Entity(DataType):
# #     pass


# # class Summary(Document):
# #     pass


# # class Sentiment(DataType):
# #     pass


# # class Synset(DataType):
# #     pass


# # class Postag(DataType):
# #     pass


# # class Chunktag(DataType):
# #     pass


# # class Tensor3(DataType):
# #     pass


# # class Tensor4(DataType):
# #     pass


# # class Flags(DataType):
# #     pass


# # class List(DataType):
# #     def __init__(self, inner):
# #         self.inner = inner
# #         # super().__init__(**inner.tags)

# #     def depth(self):
# #         if not isinstance(self.inner, List):
# #             return 1

# #         return 1 + self.inner.depth()

# #     def __conforms__(self, other):
# #         return isinstance(other, List) and conforms(self.inner, other.inner)

# #     def __repr__(self):
# #         return "List(%r)" % self.inner


# # class Tuple(DataType):
# #     def __init__(self, *inner):
# #         self.inner = inner
# #         # super().__init__(**inner[0].tags)

# #     def __repr__(self):
# #         items = ", ".join(repr(s) for s in self.inner)
# #         return "Tuple(%s)" % items

# #     def __conforms__(self, other):
# #         if not isinstance(other, Tuple):
# #             return False

# #         if len(self.inner) != len(other.inner):
# #             return False

# #         for x, y in zip(self.inner, other.inner):
# #             if not conforms(x, y):
# #                 return False

# #         return True


# # DATA_TYPES = frozenset(
# #     [
# #         Text,
# #         Word,
# #         Stem,
# #         Sentence,
# #         Document,
# #         Category,
# #         Vector,
# #         Matrix,
# #         DenseMatrix,
# #         SparseMatrix,
# #         ContinuousVector,
# #         DiscreteVector,
# #         CategoricalVector,
# #         MatrixContinuous,
# #         MatrixCategorical,
# #         MatrixContinuousDense,
# #         MatrixContinuousSparse,
# #         Entity,
# #         Summary,
# #         Sentiment,
# #         Synset,
# #         Postag,
# #         Chunktag,
# #         Tensor3,
# #         List,
# #         Tuple,
# #         Flags,
# #     ]
# # )


# # def draw_data_hierarchy(output_file):
# #     """
# #     Creates an SVG representation of the `DataType` hierarchy,
# #     for documentation purposes.
# #     """
# #     import pydot

# #     classes = list(DATA_TYPES) + [DataType]

# #     graph = pydot.Dot(direction="TB")

# #     for clss in classes:
# #         graph.add_node(pydot.Node(clss.__name__))

# #     for clss in classes:
# #         for base in clss.__bases__:
# #             if base not in classes:
# #                 continue

# #             graph.add_edge(pydot.Edge(base.__name__, clss.__name__))

# #     graph.write(output_file + ".svg", format="svg")
# #     graph.write(output_file + ".png", format="png")



def find_classes(include=None, exclude=None, modules=None, input=None, output=None):
    import inspect
    import re

    result = []

    if include:
        include = f".*({include}).*"
    else:
        include = r".*"

    if exclude:
        exclude = f".*({exclude}).*"

    if input:
        input = f".*({input}).*"

    if output:
        output = f".*({output}).*"

    if modules is None:
        modules = []

        try:
            from autogoal.contrib import sklearn

            modules.append(sklearn)
        except ImportError as e:
            pass

        try:
            from autogoal.contrib import nltk

            modules.append(nltk)
        except ImportError as e:
            pass

        try:
            from autogoal.contrib import gensim

            modules.append(gensim)
        except ImportError as e:
            pass

        try:
            from autogoal.contrib import keras

            modules.append(keras)
        except ImportError as e:
            pass

        try:
            from autogoal.contrib import torch

            modules.append(torch)
        except ImportError as e:
            pass

        try:
            from autogoal.contrib import spacy

            modules.append(spacy)
        except ImportError as e:
            pass

        try:
            from autogoal.contrib import wikipedia

            modules.append(wikipedia)
        except ImportError as e:
            pass

        from autogoal.contrib import wrappers

        modules.append(wrappers)

        from autogoal.contrib import regex

        modules.append(regex)

    for module in modules:
        for _, cls in inspect.getmembers(module, inspect.isclass):
            if not hasattr(cls, "run"):
                continue

            if cls.__name__.startswith("_"):
                continue

            if not re.match(include, repr(cls)):
                continue

            if exclude is not None and re.match(exclude, repr(cls)):
                continue

            if not cls.__module__.startswith("autogoal.contrib"):
                continue

            sig = inspect.signature(cls.run)

            if input and not re.match(input, str(sig.parameters["input"].annotation)):
                continue

            if output and not re.match(output, str(sig.return_annotation)):
                continue

            result.append(cls)

    return result


import enum


class ContribStatus(enum.Enum):
    RequiresDependency = enum.auto()
    RequiresDownload = enum.auto()
    Ready = enum.auto()


def status():
    status = {}
    modules = []

    try:
        from autogoal.contrib import sklearn

        modules.append(sklearn)
    except ImportError as e:
        status["autogoal.contrib.sklearn"] = ContribStatus.RequiresDependency

    try:
        from autogoal.contrib import nltk

        modules.append(nltk)
    except ImportError as e:
        status["autogoal.contrib.nltk"] = ContribStatus.RequiresDependency

    try:
        from autogoal.contrib import gensim

        modules.append(gensim)
    except ImportError as e:
        status["autogoal.contrib.gensim"] = ContribStatus.RequiresDependency

    try:
        from autogoal.contrib import keras

        modules.append(keras)
    except ImportError as e:
        status["autogoal.contrib.keras"] = ContribStatus.RequiresDependency

    try:
        from autogoal.contrib import torch

        modules.append(torch)
    except ImportError as e:
        status["autogoal.contrib.torch"] = ContribStatus.RequiresDependency

    try:
        from autogoal.contrib import spacy

        modules.append(spacy)
    except ImportError as e:
        status["autogoal.contrib.spacy"] = ContribStatus.RequiresDependency

    try:
        from autogoal.contrib import wikipedia

        modules.append(wikipedia)
    except ImportError as e:
        status["autogoal.contrib.wikipedia"] = ContribStatus.RequiresDependency

    for module in modules:
        if hasattr(module, "status"):
            status[module.__name__] = module.status()
        else:
            status[module.__name__] = ContribStatus.Ready

    return status


def download(contrib: str):
    modules = {}

    try:
        from autogoal.contrib import sklearn

        modules["sklearn"] = sklearn
    except ImportError as e:
        pass

    try:
        from autogoal.contrib import nltk

        modules["nltk"] = nltk
    except ImportError as e:
        pass

    try:
        from autogoal.contrib import gensim

        modules["gensim"] = gensim
    except ImportError as e:
        pass

    try:
        from autogoal.contrib import keras

        modules["keras"] = keras
    except ImportError as e:
        pass

    try:
        from autogoal.contrib import torch

        modules["torch"] = torch
    except ImportError as e:
        pass

    try:
        from autogoal.contrib import spacy

        modules["spacy"] = spacy
    except ImportError as e:
        pass

    try:
        from autogoal.contrib import wikipedia

        modules["wikipedia"] = wikipedia
    except ImportError as e:
        pass

    if contrib not in modules:
        raise ValueError(f"Contrib `{contrib}` cannot be imported.")

    contrib = modules[contrib]

    if not hasattr(contrib, "download"):
        return False

    return contrib.download()


__all__ = ["find_classes", "status", "download"]



import io
import pickle
import statistics

import numpy as np
from autogoal.contrib import find_classes

from autogoal.kb import build_pipeline_graph, SemanticType
from autogoal.ml.metrics import accuracy
from autogoal.search import PESearch
from autogoal.utils import nice_repr


@nice_repr
class AutoML:
    """
    Predefined pipeline search with automatic type inference.

    An `AutoML` instance represents a general-purpose machine learning
    algorithm, that can be applied to any input and output.
    """

    def __init__(
        self,
        input=None,
        output=None,
        random_state=None,
        search_algorithm=None,
        search_iterations=100,
        include_filter=".*",
        exclude_filter=None,
        validation_split=0.3,
        errors="warn",
        cross_validation="median",
        cross_validation_steps=3,
        registry=None,
        score_metric=None,
        **search_kwargs
    ):
        self.input = input
        self.output = output
        self.search_algorithm = search_algorithm or PESearch
        self.search_iterations = search_iterations
        self.include_filter = include_filter
        self.exclude_filter = exclude_filter
        self.validation_split = validation_split
        self.errors = errors
        self.cross_validation = cross_validation
        self.cross_validation_steps = cross_validation_steps
        self.registry = registry
        self.random_state = random_state
        self.score_metric = score_metric or accuracy
        self.search_kwargs = search_kwargs
        self._unpickled = False

        if random_state:
            np.random.seed(random_state)

    def _check_fitted(self):
        if not hasattr(self, "best_pipeline_"):
            raise TypeError(
                "This operation cannot be performed on an unfitted AutoML instance. Call `fit` first."
            )

    def make_pipeline_builder(self):
        registry = self.registry or find_classes(
            include=self.include_filter, exclude=self.exclude_filter
        )

        return build_pipeline_graph(
            input_types=self.input, output_type=self.output, registry=registry,
        )

    def fit(self, X, y, **kwargs):
        self.input = self._input_type(X)
        self.output = self._output_type(y)

        search = self.search_algorithm(
            self.make_pipeline_builder(),
            self.make_fitness_fn(X, y),
            random_state=self.random_state,
            errors=self.errors,
            **self.search_kwargs,
        )

        self.best_pipeline_, self.best_score_ = search.run(
            self.search_iterations, **kwargs
        )

        self.fit_pipeline(X, y)

    def fit_pipeline(self, X, y):
        self._check_fitted()

        self.best_pipeline_.send("train")
        self.best_pipeline_.run(X, y)
        self.best_pipeline_.send("eval")

    def save(self, fp: io.BytesIO):
        """
        Serializes the AutoML instance.
        """
        self._check_fitted()
        pickle.Pickler(fp).dump(self)

    @classmethod
    def load(self, fp: io.FileIO) -> "AutoML":
        """
        Deserializes an AutoML instance. 
        
        After deserialization, the best pipeline found is ready to predict.
        """
        automl = pickle.Unpickler(fp).load()

        if not isinstance(automl, AutoML):
            raise ValueError("The serialized file does not contain an AutoML instance.")

        return automl

    def score(self, X, y):
        self._check_fitted()

        y_pred = self.best_pipeline_.run(X, np.zeros_like(y))
        return self.score_metric(y, y_pred)

    def _input_type(self, X):
        return self.input or SemanticType.infer(X)

    def _output_type(self, y):
        return self.output or SemanticType.infer(y)

    def make_fitness_fn(self, X, y):
        y = np.asarray(y)

        def fitness_fn(pipeline):
            scores = []

            for _ in range(self.cross_validation_steps):
                len_x = len(X) if isinstance(X, list) else X.shape[0]
                indices = np.arange(0, len_x)
                np.random.shuffle(indices)
                split_index = int(self.validation_split * len(indices))
                train_indices = indices[:-split_index]
                test_indices = indices[-split_index:]

                if isinstance(X, list):
                    X_train, y_train, X_test, y_test = (
                        [X[i] for i in train_indices],
                        y[train_indices],
                        [X[i] for i in test_indices],
                        y[test_indices],
                    )
                else:
                    X_train, y_train, X_test, y_test = (
                        X[train_indices],
                        y[train_indices],
                        X[test_indices],
                        y[test_indices],
                    )

                pipeline.send("train")
                pipeline.run(X_train, y_train)
                pipeline.send("eval")
                y_pred = pipeline.run(X_test, None)
                scores.append(self.score_metric(y_test, y_pred))

            return getattr(statistics, self.cross_validation)(scores)

        return fitness_fn

    def predict(self, X):
        self._check_fitted()

        return self.best_pipeline_.run(X, None)



from sklearn.feature_extraction.text import CountVectorizer as _CountVectorizer
from sklearn.feature_extraction import DictVectorizer
from sklearn_crfsuite import CRF

from autogoal.contrib.sklearn._builder import SklearnTransformer, SklearnEstimator
from autogoal.kb import *
from autogoal.grammar import (
    BooleanValue,
    CategoricalValue,
    DiscreteValue,
    ContinuousValue,
)
from autogoal.utils import nice_repr
from autogoal.kb import AlgorithmBase, Supervised


@nice_repr
class CountVectorizerNoTokenize(_CountVectorizer, SklearnTransformer):
    def __init__(
        self,
        lowercase: BooleanValue(),
        stopwords_remove: BooleanValue(),
        binary: BooleanValue(),
        inner_tokenizer: algorithm(Sentence, Seq[Word]),
        inner_stemmer: algorithm(Word, Stem),
        inner_stopwords: algorithm(Seq[Word], Seq[Word]),
    ):
        self.stopwords_remove = stopwords_remove
        self.inner_tokenizer = inner_tokenizer
        self.inner_stemmer = inner_stemmer
        self.inner_stopwords = inner_stopwords

        SklearnTransformer.__init__(self)
        _CountVectorizer.__init__(self, lowercase=lowercase, binary=binary)

    def build_tokenizer(self):
        def func(sentence):
            tokens = self.inner_tokenizer.run(sentence)
            tokens = (
                self.inner_stopwords.run(sentence) if self.stopwords_remove else tokens
            )
            return [self.inner_stemmer.run(token) for token in tokens]

        return func

    def run(self, input: Seq[Sentence]) -> MatrixContinuousSparse:
        return SklearnTransformer.run(self, input)


class _FeatureVectorizer(SklearnTransformer):
    def __init__(self, sparse):
        self.vectorizer = DictVectorizer(sparse=sparse)
        super().__init__()

    def fit_transform(self, X, y=None):
        return self.vectorizer.fit_transform(X)

    def transform(self, X, y=None):
        return self.vectorizer.transform(X, y=y)


@nice_repr
class FeatureSparseVectorizer(_FeatureVectorizer):
    def __init__(self):
        super().__init__(sparse=True)

    def run(self, input: Seq[FeatureSet]) -> MatrixContinuousSparse:
        return super().run(input)


@nice_repr
class FeatureDenseVectorizer(_FeatureVectorizer):
    def __init__(self):
        super().__init__(sparse=False)

    def run(self, input: Seq[FeatureSet]) -> MatrixContinuousDense:
        return super().run(input)


@nice_repr
class CRFTagger(CRF, SklearnEstimator):
    def __init__(
        self, algorithm: CategoricalValue("lbfgs", "l2sgd", "ap", "pa", "arow")
    ) -> None:
        SklearnEstimator.__init__(self)
        super().__init__(algorithm=algorithm)

    def run(
        self, X: Seq[Seq[FeatureSet]], y: Supervised[Seq[Seq[Label]]]
    ) -> Seq[Seq[Label]]:
        return SklearnEstimator.run(self, X, y)


__all__ = [
    "CountVectorizerNoTokenize",
    "FeatureSparseVectorizer",
    "FeatureDenseVectorizer",
    "CRFTagger",
]



import os
import nltk
from pathlib import Path

CONTRIB_NAME = "nltk"
DATA_PATH = Path.home() / ".autogoal" / "contrib" / CONTRIB_NAME / "data"

# ensure data path directory creation
try:
    os.makedirs(DATA_PATH)
except IOError as ex:
    # directory already exists
    pass


def load_data(path):
    nltk.download(download_dir=path, quiet=True)
    use_data(path)


def use_data(path):
    nltk.data.path.append(path)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autogoal',
 'autogoal.contrib',
 'autogoal.contrib.gensim',
 'autogoal.contrib.keras',
 'autogoal.contrib.nltk',
 'autogoal.contrib.regex',
 'autogoal.contrib.sklearn',
 'autogoal.contrib.spacy',
 'autogoal.contrib.streamlit',
 'autogoal.contrib.telegram',
 'autogoal.contrib.torch',
 'autogoal.contrib.wikipedia',
 'autogoal.datasets',
 'autogoal.datasets.ehealthkd20',
 'autogoal.experimental',
 'autogoal.grammar',
 'autogoal.kb',
 'autogoal.logging',
 'autogoal.ml',
 'autogoal.sampling',
 'autogoal.search',
 'autogoal.utils']

package_data = \
{'': ['*']}

install_requires = \
['black>=19.10b0,<20.0',
 'enlighten>=1.4.0,<2.0.0',
 'networkx>=2.4,<3.0',
 'numpy>=1.19.2,<2.0.0',
 'pandas>=1.1.3,<2.0.0',
 'psutil>=5.6.7,<6.0.0',
 'pydot>=1.4.1,<2.0.0',
 'pyyaml>=5.2,<6.0',
 'rich>=8.0.0,<9.0.0',
 'scipy>=1.5.2,<2.0.0',
 'termcolor>=1.1.0,<2.0.0',
 'toml>=0.10.0,<0.11.0',
 'tqdm>=4.50.2,<5.0.0',
 'typer>=0.3.2,<0.4.0']

extras_require = \
{'contrib': ['gensim>=3.8.1,<4.0.0',
             'jupyterlab>=1.2.4,<2.0.0',
             'keras>=2.3.1,<3.0.0',
             'nltk>=3.4.5,<4.0.0',
             'nx_altair>=0.1.4,<0.2.0',
             'python-telegram-bot>=12.4.2,<13.0.0',
             'scikit-learn>=0.22,<0.23',
             'seqlearn>=0.2,<0.3',
             'sklearn_crfsuite>=0.3.6,<0.4.0',
             'spacy>=2.2.3,<3.0.0',
             'streamlit>=0.59.0,<0.60.0',
             'transformers>=2.3.0,<3.0.0',
             'wikipedia>=1.4.0,<2.0.0'],
 'dev': ['codecov>=2.0.15,<3.0.0',
         'markdown-include>=0.5.1,<0.6.0',
         'mkdocs>=1.0.4,<2.0.0',
         'mkdocs-material>=4.6.0,<5.0.0',
         'mypy>=0.761,<0.762',
         'pylint>=2.4.4,<3.0.0',
         'pytest>=5.3.2,<6.0.0',
         'pytest-cov>=2.8.1,<3.0.0',
         'typer-cli>=0.0.11,<0.0.12'],
 'gensim': ['gensim>=3.8.1,<4.0.0'],
 'keras': ['keras>=2.3.1,<3.0.0'],
 'nltk': ['nltk>=3.4.5,<4.0.0'],
 'sklearn': ['scikit-learn>=0.22,<0.23',
             'seqlearn>=0.2,<0.3',
             'sklearn_crfsuite>=0.3.6,<0.4.0'],
 'spacy': ['spacy>=2.2.3,<3.0.0'],
 'streamlit': ['nx_altair>=0.1.4,<0.2.0', 'streamlit>=0.59.0,<0.60.0'],
 'telegram': ['python-telegram-bot>=12.4.2,<13.0.0'],
 'wikipedia': ['wikipedia>=1.4.0,<2.0.0']}

setup_kwargs = {
    'name': 'autogoal',
    'version': '0.4.3',
    'description': 'Automatic Generation Optimization And Learning',
    'long_description': '![AutoGOAL Logo](https://autogoal.github.io/autogoal-banner.png)\n\n[<img alt="PyPI" src="https://img.shields.io/pypi/v/autogoal">](https://pypi.org/project/autogoal/) [<img alt="PyPI - Python Version" src="https://img.shields.io/pypi/pyversions/autogoal">](https://pypi.org/project/autogoal/) [<img alt="PyPI - License" src="https://img.shields.io/pypi/l/autogoal">](https://autogoal.github.io/contributing) [<img alt="GitHub stars" src="https://img.shields.io/github/stars/autogoal/autogoal?style=social">](https://github.com/autogoal/autogoal/stargazers) [<img alt="Twitter Follow" src="https://img.shields.io/twitter/follow/auto_goal?label=Followers&style=social">](https://twitter.com/auto_goal)\n\n[<img alt="GitHub Workflow Status (branch)" src="https://img.shields.io/github/workflow/status/autogoal/autogoal/CI/main?label=unit tests&logo=github">](https://github.com/autogoal/autogoal/actions)\n[<img src="https://codecov.io/gh/autogoal/autogoal/branch/main/graph/badge.svg" />](https://codecov.io/gh/autogoal/autogoal/)\n[<img alt="Docker Image Size (CPU)" src="https://img.shields.io/docker/image-size/autogoal/autogoal/latest">](https://hub.docker.com/r/autogoal/autogoal)\n[<img alt="Docker Pulls" src="https://img.shields.io/docker/pulls/autogoal/autogoal">](https://hub.docker.com/r/autogoal/autogoal)\n\n# AutoGOAL\n\n> Automatic Generation, Optimization And Artificial Learning\n\nAutoGOAL is a Python library for automatically finding the best way to solve a given task.\nIt has been designed mainly for _Automated Machine Learning_ (aka [AutoML](https://www.automl.org))\nbut it can be used in any scenario where you have several possible ways to solve a given task.\n\nTechnically speaking, AutoGOAL is a framework for program synthesis, i.e., finding the best program to solve\na given problem, provided that the user can describe the space of all possible programs.\nAutoGOAL provides a set of low-level components to define different spaces and efficiently search in them.\nIn the specific context of machine learning, AutoGOAL also provides high-level components that can be used as a black-box in almost any type of problem and dataset format.\n\n## ⭐ Quickstart\n\nAutoGOAL is first and foremost a framework for Automated Machine Learning.\nAs such, it comes pre-packaged with hundreds of low-level machine learning\nalgorithms that can be automatically assembled into pipelines for different problems.\n\nThe core of this functionality lies in the [`AutoML`](https://autogoal.github.io/api/autogoal.ml#automl) class.\n\nTo illustrate the simplicity of its use we will load a dataset and run an automatic classifier in it.\nThe following code will run for approximately 5 minutes on a classic dataset.\n\n```python\nfrom autogoal.datasets import cars\nfrom autogoal.kb import (MatrixContinuousDense, \n                         Supervised, \n                         VectorCategorical)\nfrom autogoal.ml import AutoML\n\n# Load dataset\nX, y = cars.load()\n\n# Instantiate AutoML and define input/output types\nautoml = AutoML(\n    input=(MatrixContinuousDense, \n           Supervised[VectorCategorical]),\n    output=VectorCategorical\n)\n\n# Run the pipeline search process\nautoml.fit(X, y)\n\n# Report the best pipeline\nprint(automl.best_pipeline_)\nprint(automl.best_score_)\n```\n\nSensible defaults are defined for each of the many parameters of `AutoML`.\nMake sure to [read the documentation](https://autogoal.github.io/guide/) for more information.\n\n## ⚙️ Installation\n\nThe easiest way to get AutoGOAL up and running with all the dependencies is to pull the development Docker image, which is somewhat big:\n\n    docker pull autogoal/autogoal\n\nInstructions for setting up Docker are available [here](https://www.docker.com/get-started).\n\nOnce you have the development image downloaded, you can fire up a console and use AutoGOAL interactively.\n\n![](https://autogoal.github.io/shell.svg)\n\nIf you prefer to not use Docker, or you don\'t want all the dependencies, you can also install AutoGOAL directly with pip:\n\n    pip install autogoal\n\nThis will install the core library but you won\'t be able to use any of the underlying machine learning algorithms until you install the corresponding optional dependencies. You can install them all with:\n\n    pip install autogoal[contrib]\n\nTo fine-pick which dependencies you want, read the [dependencies section](https://autogoal.github.io/dependencies/).\n\n> ⚠️ **NOTE**: By installing through `pip` you will get the latest release version of AutoGOAL, while by installing through Docker, you will get the latest development version. \n>\n> The development version is mostly up-to-date with the `main` branch, hence it will probably contain more features, but also more bugs, than the release version.\n\n## 💻 CLI\n\nYou can use AutoGOAL directly from the CLI. To see options just type:\n\n    autogoal\n\nUsing the CLI you can train and use AutoML models, download datasets and inspect the contrib libraries without writing a single line of code.\n\n![](https://autogoal.github.io/shell/autogoal_cli.svg)\n\nRead more in the [CLI documentation](https://autogoal.github.io/cli).\n\n## \U0001f929 Demo\n\nAn online demo app is available at [autogoal.github.io/demo](https://autogoal.github.io/demo).\nThis app showcases the main features of AutoGOAL in interactive case studies.\n\nTo run the demo locally, simply type:\n\n    docker run -p 8501:8501 autogoal/autogoal\n\nAnd navigate to [localhost:8501](http://localhost:8501).\n\n## ⚖️ API stability\n\nWe make a conscious effort to maintain a consistent public API across versions, but the private API can change at any time.\nIn general, everything you can import from `autogoal` without underscores is considered public. \n\nFor example:\n\n```python\n# "clean" imports are part of the public API\nfrom autogoal import optimize   \nfrom autogoal.ml import AutoML  \nfrom autogoal.contrib.sklearn import find_classes\n\n# public members of public types as well\nautoml = AutoML\nautoml.fit(...) \n\n# underscored imports are part of the private API\nfrom autogoal.ml._automl import ...\nfrom autogoal.contrib.sklearn._generated import ...\n\n# as well as private members of any type\nautoml._input_type(...)\n\n```\n\nThese are our consistency rules:\n\n- Major breaking changes are introduced between major version updates, e.g., `x.0` and `y.0`. These can be additions, removals, or modifications of any kind in any part of the API.\n\n- Between minor version updates, e.g., `1.x` and `1.y`, you can expect to find new functionality, but anything you can use from the   public API will still be there with a consistent semantic (save for bugfixes).\n\n- Between micro version updates, e.g., `1.3.x` and `1.3.y`, the public API is frozen even for additions.\n\n- The private API can be changed at all times.\n\n⚠️ While AutoGOAL is on public beta (versions `0.x`) the public API is considered unstable and thus everything can change. However, we try to keep breaking changes to a minimum.\n\n## 📚 Documentation\n\nThis documentation is available online at [autogoal.github.io](https://autogoal.github.io). Check the following sections:\n\n- [**User Guide**](https://autogoal.github.io/guide/): Step-by-step showcase of everything you need to know to use AuoGOAL.\n- [**Examples**](https://autogoal.github.io/examples/): The best way to learn how to use AutoGOAL by practice.\n- [**API**](https://autogoal.github.io/api/autogoal): Details about the public API for AutoGOAL.\n\nThe HTML version can be deployed offline by downloading the [AutoGOAL Docker image](https://hub.docker.com/autogoal/autogoal) and running:\n\n    docker run -p 8000:8000 autogoal/autogoal mkdocs serve -a 0.0.0.0:8000\n\nAnd navigating to [localhost:8000](http://localhost:8000).\n\n## 📃 Publications\n\nIf you use AutoGOAL in academic research, please cite the following paper:\n\n```bibtex\n@article{estevez2020general,\n  title={General-purpose hierarchical optimisation of machine learning pipelines with grammatical evolution},\n  author={Est{\\\'e}vez-Velarde, Suilan and Guti{\\\'e}rrez, Yoan and Almeida-Cruz, Yudivi{\\\'a}n and Montoyo, Andr{\\\'e}s},\n  journal={Information Sciences},\n  year={2020},\n  publisher={Elsevier},\n  doi={10.1016/j.ins.2020.07.035}\n}\n```\n\nThe technologies and theoretical results leading up to AutoGOAL have been presented at different venues:\n\n- [Optimizing Natural Language Processing Pipelines: Opinion Mining Case Study](https://link.springer.com/chapter/10.1007/978-3-030-33904-3_15) marks the inception of the idea of using evolutionary optimization with a probabilistic search space for pipeline optimization.\n\n- [AutoML Strategy Based on Grammatical Evolution: A Case Study about Knowledge Discovery from Text](https://www.aclweb.org/anthology/P19-1428/) applied probabilistic grammatical evolution with a custom-made grammar in the context of entity recognition in medical text.\n\n- [General-purpose Hierarchical Optimisation of Machine Learning Pipelines with Grammatical Evolution](https://doi.org/10.1016/j.ins.2020.07.035) presents a more uniform framework with different grammars in different problems, from tabular datasets to natural language processing.\n\n- [Solving Heterogeneous AutoML Problems with AutoGOAL](https://www.automl.org/wp-content/uploads/2020/07/AutoML_2020_paper_20.pdf) is the first actual description of AutoGOAL as a framework, unifying the ideas presented in the previous papers.\n\n## 🤝 Contribution\n\nCode is licensed under MIT. Read the details in the [collaboration section](https://autogoal.github.io/contributing).\n\nThis project follows the [all-contributors](https://allcontributors.org) specification. Any contribution will be given credit, from fixing typos, to reporting bugs, to implementing new core functionalities. \n\nHere are all the current contributions. \n\n> **🙏 Thanks!**\n\n<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->\n<!-- prettier-ignore-start -->\n<!-- markdownlint-disable -->\n<table>\n  <tr>\n    <td align="center"><a href="https://github.com/sestevez"><img src="https://avatars3.githubusercontent.com/u/6156391?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Suilan Estevez-Velarde</b></sub></a><br /><a href="https://github.com/autogoal/autogoal/commits?author=sestevez" title="Code">💻</a> <a href="https://github.com/autogoal/autogoal/commits?author=sestevez" title="Tests">⚠️</a> <a href="#ideas-sestevez" title="Ideas, Planning, & Feedback">🤔</a> <a href="https://github.com/autogoal/autogoal/commits?author=sestevez" title="Documentation">📖</a></td>\n    <td align="center"><a href="https://apiad.net"><img src="https://avatars3.githubusercontent.com/u/1778204?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Alejandro Piad</b></sub></a><br /><a href="https://github.com/autogoal/autogoal/commits?author=apiad" title="Code">💻</a> <a href="https://github.com/autogoal/autogoal/commits?author=apiad" title="Tests">⚠️</a> <a href="https://github.com/autogoal/autogoal/commits?author=apiad" title="Documentation">📖</a></td>\n    <td align="center"><a href="https://github.com/yudivian"><img src="https://avatars1.githubusercontent.com/u/5324359?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Yudivián Almeida Cruz</b></sub></a><br /><a href="#ideas-yudivian" title="Ideas, Planning, & Feedback">🤔</a> <a href="https://github.com/autogoal/autogoal/commits?author=yudivian" title="Documentation">📖</a></td>\n    <td align="center"><a href="http://orcid.org/0000-0002-4052-7427"><img src="https://avatars2.githubusercontent.com/u/25705914?v=4?s=100" width="100px;" alt=""/><br /><sub><b>ygutierrez</b></sub></a><br /><a href="#ideas-joogvzz" title="Ideas, Planning, & Feedback">🤔</a> <a href="https://github.com/autogoal/autogoal/commits?author=joogvzz" title="Documentation">📖</a></td>\n    <td align="center"><a href="https://github.com/EEstevanell"><img src="https://avatars0.githubusercontent.com/u/45082075?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Ernesto Luis Estevanell Valladares</b></sub></a><br /><a href="https://github.com/autogoal/autogoal/commits?author=EEstevanell" title="Code">💻</a> <a href="https://github.com/autogoal/autogoal/commits?author=EEstevanell" title="Tests">⚠️</a></td>\n    <td align="center"><a href="http://alexfertel.netlify.app"><img src="https://avatars3.githubusercontent.com/u/22298999?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Alexander Gonzalez</b></sub></a><br /><a href="https://github.com/autogoal/autogoal/commits?author=alexfertel" title="Code">💻</a> <a href="https://github.com/autogoal/autogoal/commits?author=alexfertel" title="Tests">⚠️</a></td>\n    <td align="center"><a href="https://www.linkedin.com/in/anshu-trivedi-501a7b146/"><img src="https://avatars1.githubusercontent.com/u/47869948?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Anshu Trivedi</b></sub></a><br /><a href="https://github.com/autogoal/autogoal/commits?author=AnshuTrivedi" title="Code">💻</a></td>\n  </tr>\n  <tr>\n    <td align="center"><a href="http://alxrcs.github.io"><img src="https://avatars1.githubusercontent.com/u/8171561?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Alex Coto</b></sub></a><br /><a href="https://github.com/autogoal/autogoal/commits?author=alxrcs" title="Documentation">📖</a></td>\n    <td align="center"><a href="https://github.com/geblanco"><img src="https://avatars3.githubusercontent.com/u/6652222?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Guillermo Blanco</b></sub></a><br /><a href="https://github.com/autogoal/autogoal/issues?q=author%3Ageblanco" title="Bug reports">🐛</a> <a href="https://github.com/autogoal/autogoal/commits?author=geblanco" title="Code">💻</a> <a href="https://github.com/autogoal/autogoal/commits?author=geblanco" title="Documentation">📖</a></td>\n    <td align="center"><a href="https://github.com/yacth"><img src="https://avatars3.githubusercontent.com/u/71322097?v=4?s=100" width="100px;" alt=""/><br /><sub><b>yacth</b></sub></a><br /><a href="https://github.com/autogoal/autogoal/issues?q=author%3Ayacth" title="Bug reports">🐛</a> <a href="https://github.com/autogoal/autogoal/commits?author=yacth" title="Code">💻</a></td>\n    <td align="center"><a href="https://sourceplusplus.com"><img src="https://avatars0.githubusercontent.com/u/3278877?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Brandon Fergerson</b></sub></a><br /><a href="https://github.com/autogoal/autogoal/issues?q=author%3ABFergerson" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://adityanikhil.github.io/main/"><img src="https://avatars2.githubusercontent.com/u/30192967?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Aditya Nikhil</b></sub></a><br /><a href="https://github.com/autogoal/autogoal/issues?q=author%3AAdityaNikhil" title="Bug reports">🐛</a></td>\n  </tr>\n</table>\n\n<!-- markdownlint-restore -->\n<!-- prettier-ignore-end -->\n\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n \n',
    'author': 'Suilan Estevez-Velarde',
    'author_email': 'suilanestevez@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://autogoal.github.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



import logging
import warnings

from rich.logging import RichHandler
from rich.console import Console


_CONSOLE = Console()


def setup(level="INFO"):
    logging.basicConfig(
        level=level,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True, console=_CONSOLE)],
    )
    logging.captureWarnings(True)


def logger() -> logging.Logger:
    return logging.getLogger("autogoal")


def console() -> Console:
    return _CONSOLE



import shutil
from pathlib import Path
from typing import Dict
import json

import requests
import os
from tqdm import tqdm
from functools import lru_cache


DATASETS_METADATA = (
    "https://raw.githubusercontent.com/autogoal/datasets/master/datasets.json"
)

DATA_PATH = Path.home() / ".autogoal" / "data"

# ensure data path directory creation
os.makedirs(DATA_PATH, exist_ok=True)


@lru_cache()
def get_datasets_list() -> Dict[str, str]:
    try:
        data = requests.get(DATASETS_METADATA).json()

        with open(DATA_PATH / "datasets.json", "w") as fp:
            json.dump(data, fp, indent=2)

        return data
    except requests.ConnectionError as e:
        try:
            with open(DATA_PATH / "datasets.json", "r") as fp:
                return json.load(fp)
        except IOError:
            raise Exception(
                "Cannot download dataset list and no cached version exists."
            )


def datapath(path: str) -> Path:
    """
    Returns a `Path` object that points to the dataset path
    where `path` is located.

    ##### Examples

    ```python
    >>> datapath("movie_reviews")
    PosixPath('/home/coder/autogoal/autogoal/datasets/data/movie_reviews')

    ```
    """
    return Path(DATA_PATH) / path


def pack(folder: str):
    filename = datapath(folder)
    rootdir = datapath(folder)
    shutil.make_archive(filename, "zip", root_dir=rootdir)


def unpack(zipfile: str):
    filename = datapath(zipfile)
    rootdir = datapath(zipfile[:-4])
    shutil.unpack_archive(filename, extract_dir=rootdir, format="zip")


def download(dataset: str, unpackit: bool = True):
    fname = f"{dataset}.zip"
    path = datapath(fname)

    if path.exists():
        return

    datasets = get_datasets_list()
    url = datasets[dataset]

    download_and_save(url, path, True)

    if unpackit:
        unpack(fname)


def download_and_save(url, path: Path, overwrite=False, data_length=None):
    stream = requests.get(url, stream=True)
    total_size = data_length or int(stream.headers.get("content-length", 0))

    if path.exists() and not overwrite:
        return False

    try:
        with path.open("wb") as f:
            with tqdm(
                total=total_size, unit="B", unit_scale=True, unit_divisor=1024
            ) as pbar:
                for data in stream.iter_content(32 * 1024):
                    f.write(data)
                    pbar.update(len(data))

        return True
    except:
        path.unlink()
        raise



from autogoal.kb import AlgorithmBase, Supervised
from typing import Optional

import collections
import numpy as np
from tensorflow.keras.callbacks import EarlyStopping, TerminateOnNaN
from tensorflow.keras.layers import Dense, Input, TimeDistributed, concatenate
from tensorflow.keras.models import Model
from tensorflow.keras.utils import to_categorical
from tensorflow.keras.preprocessing.image import (
    ImageDataGenerator as _ImageDataGenerator,
)

from autogoal.contrib.keras._grammars import build_grammar, generate_grammar, Modules
from autogoal.grammar import (
    Graph,
    GraphGrammar,
    Sampler,
    BooleanValue,
    CategoricalValue,
    ContinuousValue,
    DiscreteValue,
)
from autogoal.kb import (
    VectorCategorical,
    Seq,
    MatrixContinuousDense,
    Postag,
    Tensor3,
    Tensor4,
)

from autogoal.utils import nice_repr
import abc


@nice_repr
class KerasNeuralNetwork(AlgorithmBase, metaclass=abc.ABCMeta):
    def __init__(
        self,
        grammar: GraphGrammar,
        optimizer: CategoricalValue("sgd", "adam", "rmsprop"),
        epochs=10,
        early_stop=3,
        validation_split=0.1,
        **compile_kwargs,
    ):
        self.optimizer = optimizer
        self._grammar = grammar
        self._epochs = epochs
        self._compile_kwargs = compile_kwargs
        self._model: Optional[Model] = None
        self._mode = "train"
        self._graph = None
        self._validation_split = validation_split
        self._early_stop = early_stop

    def train(self):
        self._mode = "train"

    def eval(self):
        self._mode = "eval"

    def run(self, X, y=None):
        if self._mode == "train":
            self.fit(X, y)
            return y
        if self._mode == "eval":
            return self.predict(X)

        assert False, "Invalid mode %s" % self._mode

    def __repr__(self):
        nodes = len(self._graph.nodes) if self._graph is not None else None
        return f"{self.__class__.__name__}(nodes={nodes}, compile_kwargs={self._compile_kwargs})"

    def __nice_repr_hook__(self, names, values):
        graph = []

        if self._graph is not None:
            for node, _ in self._graph.build_order():
                graph.append(node)

        names.append("graph")
        values.append(graph)

    @property
    def model(self):
        if self._model is None:
            raise TypeError("You need to call `fit` first to generate the model.")

        return self._model

    def sample(self, sampler: Sampler = None, max_iterations=100):
        if sampler is None:
            sampler = Sampler()

        self._graph = self._grammar.sample(
            sampler=sampler, max_iterations=max_iterations
        )
        return self

    def _build_nn(self, graph: Graph, X, y):
        input_x = self._build_input(X)

        def build_model(layer, _, previous_layers):
            if not previous_layers:
                return layer(input_x)

            if len(previous_layers) > 1:
                incoming = concatenate(previous_layers)
            else:
                incoming = previous_layers[0]

            return layer(incoming)

        output_y = graph.apply(build_model) or [input_x]
        final_ouput = self._build_output(output_y, y)

        if "optimizer" not in self._compile_kwargs:
            self._compile_kwargs["optimizer"] = self.optimizer

        self._model = Model(inputs=input_x, outputs=final_ouput)
        self._model.compile(**self._compile_kwargs)

    @abc.abstractmethod
    def _build_input(self, X):
        pass

    def _build_output(self, outputs, y):
        return outputs

    def fit(self, X, y, **kwargs):
        if self._graph is None:
            raise TypeError("You must call `sample` to generate the internal model.")

        self._build_nn(self._graph, X, y)
        # self.model.summary()
        self._fit_model(X, y, **kwargs)

    def _fit_model(self, X, y, **kwargs):
        self.model.fit(
            x=X,
            y=y,
            epochs=self._epochs,
            callbacks=[
                EarlyStopping(patience=self._early_stop, restore_best_weights=True),
                TerminateOnNaN(),
            ],
            validation_split=self._validation_split,
            verbose=0,
            **kwargs,
        )

    def predict(self, X):
        return self.model.predict(X)


class KerasClassifier(KerasNeuralNetwork):
    def __init__(
        self,
        optimizer: CategoricalValue("sgd", "adam", "rmsprop"),
        grammar=None,
        **kwargs,
    ):
        self._classes = None
        self._num_classes = None
        super().__init__(
            grammar=grammar or self._build_grammar(), optimizer=optimizer, **kwargs
        )

    def _build_grammar(self):
        return build_grammar(features=True)

    def _build_input(self, X):
        return Input(shape=(X.shape[1],))

    def _build_output_layer(self, y):
        self._num_classes = y.shape[1]

        if "loss" not in self._compile_kwargs:
            self._compile_kwargs["loss"] = "categorical_crossentropy"
            self._compile_kwargs["metrics"] = ["accuracy"]

        return Dense(units=self._num_classes, activation="softmax")

    def _build_output(self, outputs, y):
        if len(outputs) > 1:
            outputs = concatenate(outputs)
        else:
            outputs = outputs[0]

        return self._build_output_layer(y)(outputs)

    def fit(self, X, y):
        self._classes = {k: v for k, v in zip(set(y), range(len(y)))}
        self._inverse_classes = {v: k for k, v in self._classes.items()}
        y = [self._classes[yi] for yi in y]
        y = to_categorical(y)
        return super(KerasClassifier, self).fit(X, y)

    def predict(self, X):
        if self._classes is None:
            raise TypeError(
                "You must call `fit` before `predict` to learn class mappings."
            )

        predictions = super(KerasClassifier, self).predict(X)
        predictions = predictions.argmax(axis=-1)
        return [self._inverse_classes[yi] for yi in predictions]

    def run(
        self, X: MatrixContinuousDense, y: Supervised[VectorCategorical]
    ) -> VectorCategorical:
        return super().run(input)


@nice_repr
class KerasImagePreprocessor(_ImageDataGenerator):
    """Augment a dataset of images by making changes to the original training set.

    Applies standard dataset augmentation strategies, such as rotating,
    scaling and fliping the image.
    Uses the `ImageDataGenerator` class from keras.

    The parameter `grow_size` determines how many new images will be created for each original image.
    The remaining parameters are passed to `ImageDataGenerator`.
    """

    def __init__(
        self,
        featurewise_center: BooleanValue(),
        samplewise_center: BooleanValue(),
        featurewise_std_normalization: BooleanValue(),
        samplewise_std_normalization: BooleanValue(),
        rotation_range: DiscreteValue(0, 15),
        width_shift_range: ContinuousValue(0, 0.25),
        height_shift_range: ContinuousValue(0, 0.25),
        shear_range: ContinuousValue(0, 15),
        zoom_range: ContinuousValue(0, 0.25),
        horizontal_flip: BooleanValue(),
        vertical_flip: BooleanValue(),
    ):
        super().__init__(
            featurewise_center=featurewise_center,
            samplewise_center=samplewise_center,
            featurewise_std_normalization=featurewise_std_normalization,
            samplewise_std_normalization=samplewise_std_normalization,
            rotation_range=rotation_range,
            width_shift_range=width_shift_range,
            height_shift_range=height_shift_range,
            shear_range=shear_range,
            zoom_range=zoom_range,
            horizontal_flip=horizontal_flip,
            vertical_flip=vertical_flip,
        )


class KerasImageClassifier(KerasClassifier):
    def __init__(
        self,
        preprocessor: KerasImagePreprocessor,
        optimizer: CategoricalValue("sgd", "adam", "rmsprop"),
        **kwargs,
    ):
        self.preprocessor = preprocessor
        super().__init__(optimizer=optimizer, **kwargs)

    def _build_grammar(self):
        return generate_grammar(
            Modules.Preprocessing.Conv2D(),
            # Modules.Features.Dense()
        )

    def _fit_model(self, X, y, **kwargs):
        self.preprocessor.fit(X)
        batch_size = 64
        validation_size = min(int(0.25 * len(X)), 10 * batch_size)

        Xtrain, Xvalid = X[:-validation_size], X[-validation_size:]
        ytrain, yvalid = y[:-validation_size], y[-validation_size:]

        self.model.fit(
            self.preprocessor.flow(Xtrain, ytrain, batch_size=batch_size),
            steps_per_epoch=len(Xtrain) // batch_size,
            epochs=self._epochs,
            callbacks=[
                EarlyStopping(patience=self._early_stop, restore_best_weights=True),
                TerminateOnNaN(),
            ],
            validation_data=(Xvalid, yvalid),
            verbose=0,
            **kwargs,
        )

    def run(self, X: Tensor4, y: Supervised[VectorCategorical]) -> VectorCategorical:
        return super().run(input)

    def _build_input(self, X):
        return Input(shape=X.shape[1:])


class KerasSequenceClassifier(KerasClassifier):
    def _build_grammar(self):
        return build_grammar(preprocessing=True, reduction=True, features=True)

    def _build_input(self, X):
        return Input(shape=(None, X.shape[2]))

    def run(self, X: Tensor3, y: Supervised[VectorCategorical]) -> VectorCategorical:
        return super().run(input)


# !!! warning
# These imports are taken from <https://github.com/tensorflow/addons/pull/377>
# since the CRF layer has not yet landed in tensorflow
# TODO: Make sure to replace this when tensorflow merges this commit
from autogoal.contrib.keras._crf import CRF
from autogoal.contrib.keras._crf import crf_loss


class KerasSequenceTagger(KerasNeuralNetwork):
    def __init__(
        self,
        decode: CategoricalValue("dense", "crf"),
        optimizer: CategoricalValue("sgd", "adam", "rmsprop"),
        grammar=None,
        **kwargs,
    ):
        self._classes = None
        self._num_classes = None

        if decode not in ["dense", "crf"]:
            raise ValueError(f"Invalid decode={decode}")

        self.decode = decode
        super().__init__(
            grammar=grammar or self._build_grammar(), optimizer=optimizer, **kwargs
        )

    def _build_grammar(self):
        from autogoal.grammar._graph import Epsilon

        return GraphGrammar(Epsilon())

    def _build_input(self, X):
        return Input(shape=(None, X[0].shape[-1]))

    def _build_output(self, outputs, y):
        if "loss" not in self._compile_kwargs:
            self._compile_kwargs["metrics"] = ["accuracy"]

            if self.decode == "dense":
                self._compile_kwargs["loss"] = "categorical_crossentropy"
            elif self.decode == "crf":
                self._compile_kwargs["loss"] = crf_loss

        if len(outputs) > 1:
            outputs = concatenate(outputs)
        else:
            outputs = outputs[0]

        if self.decode == "dense":
            dense = Dense(units=len(self._classes), activation="softmax")
            return TimeDistributed(dense)(outputs)
        elif self.decode == "crf":
            crf = CRF(units=len(self._classes))
            return crf(outputs)

    def fit(self, X, y):
        distinct_classes = set(x for yi in y for x in yi)

        self._classes = {
            k: v for k, v in zip(distinct_classes, range(len(distinct_classes)))
        }
        self._inverse_classes = {v: k for k, v in self._classes.items()}

        y = [[self._classes[x] for x in yi] for yi in y]
        return super().fit(X, y)

    def _encode(self, xi, yi):
        if self.decode == "dense":
            yi = to_categorical(yi, len(self._classes))

        return (
            np.expand_dims(xi, axis=0),
            np.expand_dims(yi, axis=0),
        )

    def _generate_batches(self, X, y):
        while True:
            for xi, yi in zip(X, y):
                yield self._encode(xi, yi)

    def _fit_model(self, X, y, **kwargs):
        self.model.fit(
            self._generate_batches(X, y),
            steps_per_epoch=len(X),
            epochs=self._epochs,
            callbacks=[
                EarlyStopping(
                    patience=self._early_stop,
                    restore_best_weights=True,
                    monitor="accuracy",
                ),
                TerminateOnNaN(),
            ],
            verbose=0,
            **kwargs,
        )

    def _decode(self, predictions):
        if self.decode == "dense":
            predictions = [pr.argmax(axis=-1) for pr in predictions]
            return [[self._inverse_classes[x] for x in yi[0]] for yi in predictions]
        elif self.decode == "crf":
            return [[self._inverse_classes[x] for x in yi[0]] for yi in predictions]

    def predict(self, X):
        if self._classes is None:
            raise TypeError(
                "You must call `fit` before `predict` to learn class mappings."
            )

        predictions = [self.model.predict(np.expand_dims(xi, axis=0)) for xi in X]
        return self._decode(predictions)

    def run(
        self, X: Seq[MatrixContinuousDense], y: Supervised[Seq[Seq[Postag]]]
    ) -> Seq[Seq[Postag]]:
        return super().run(X, y)



import collections
import inspect
import logging
from os import stat
from pathlib import Path
from typing import List

import pandas as pd
import typer
from rich.console import Console
from rich.logging import RichHandler
from rich.table import Table

from autogoal.contrib import (
    find_classes,
    status,
    ContribStatus,
    download as download_contrib,
)
from autogoal.kb import VectorCategorical
from autogoal.ml import AutoML
from autogoal.search import RichLogger
from autogoal.utils import Gb, Min
from autogoal.datasets import datapath, get_datasets_list, download, dummy
import autogoal.logging

autogoal.logging.setup("WARNING")

logger = autogoal.logging.logger()
console = Console()


app = typer.Typer(name="autogoal")
contrib_app = typer.Typer(name="contrib")
automl_app = typer.Typer(name="ml")
data_app = typer.Typer(name="data")


app.add_typer(contrib_app)
app.add_typer(automl_app)
app.add_typer(data_app)


@app.callback()
def main():
    """
    🤩 Manage AutoGOAL directly from the CLI.
    """


@app.command()
def demo():
    """
    🌟 Launch streamlit demo.
    """

    try:
        from streamlit.bootstrap import run

        run(Path(__file__).parent / "contrib" / "streamlit" / "demo.py", "", "")
    except ImportError:
        console.print("(!) Too run the demo you need streamlit installed.")
        console.print("(!) Fix it by running `pip install autogoal[streamlit]`.")


@contrib_app.callback()
def contrib_main():
    """
    🔍 Inspect contrib libraries and algorithms.
    """


@contrib_app.command("list")
def contrib_list(
    verbose: bool = False,
    include: str = None,
    exclude: str = None,
    input: str = None,
    output: str = None,
):
    """
    ⚙️ List all currently available contrib algorithms.
    """
    classes = find_classes(include=include, exclude=exclude, input=input, output=output)
    classes_by_contrib = collections.defaultdict(list)
    max_cls_name_length = 0

    for cls in classes:
        max_cls_name_length = max(max_cls_name_length, len(cls.__name__))
        classes_by_contrib[str(cls).split(".")[2]].append(cls)

    typer.echo(
        f"⚙️  Found a total of {len(classes)} matching algorithms.", color="blue"
    )

    for contrib, clss in classes_by_contrib.items():
        typer.echo(f"🛠️  {contrib}: {len(clss)} algorithms.")

        if verbose:
            for cls in clss:
                sig = inspect.signature(cls.run)
                typer.echo(
                    f" 🔹 {cls.__name__.ljust(max_cls_name_length)} : {sig.parameters['input'].annotation} -> {sig.return_annotation}"
                )


@contrib_app.command("status")
def contrib_status():
    """
    ✔️ Shows the status of all contrib libraries.
    """
    table = Table("🛠️  Contrib", "✔️ Status")

    statuses = {
        ContribStatus.RequiresDependency: "🔴 Required dependency",
        ContribStatus.RequiresDownload: "🔴 Requires download",
        ContribStatus.Ready: "🟢 Ready",
    }

    for key, value in status().items():
        table.add_row(key, statuses[value])

    console.print(table)


@contrib_app.command("download")
def contrib_download(
    contrib=typer.Argument(
        ..., help="Name of the contrib, e.g., `sklearn` or `nltk`, or `all`."
    )
):
    """
    💾 Download necessary contrib files.
    """
    if status()[f"autogoal.contrib.{contrib}"] == ContribStatus.Ready:
        console.print(f"✅ Nothing to download for contrib `{contrib}`.")
    elif download_contrib(contrib):
        console.print(f"✅ Succesfully downloaded files for contrib `{contrib}`.")
    else:
        console.print(f"❌ Cannot download files for contrib `{contrib}`.")


@automl_app.callback()
def automl_callback():
    """
    🤖 Fit and predict with an AutoML model.
    """


def _load_dataset(format, input, ignore):
    if format is None:
        if input.suffix == ".csv":
            format = "csv"
        if input.suffix == ".json":
            format = "json"

    if format == "csv":
        dataset = pd.read_csv(input)
    elif format == "json":
        dataset = pd.read_json(input)
    else:
        raise ValueError("Input format not recognized. Must be either CSV or JSON.")

    if ignore:
        columns_to_ignore = [dataset.columns[i] for i in ignore]
        dataset = dataset.drop(columns=columns_to_ignore)

    return dataset


@automl_app.command("fit")
def automl_fit(
    input: Path,
    output: Path = Path("automl.bin"),
    target: str = None,
    ignore_cols: List[int] = typer.Option([]),
    evaluation_timeout: int = 5 * Min,
    memory_limit: int = 4 * Gb,
    search_timeout: int = 60 * 60,
    pop_size: int = 20,
    iterations: int = 100,
    random_state: int = None,
    format: str = None,
):
    """
    🏃 Train an AutoML instance on a dataset.
    """

    try:
        dataset = _load_dataset(format, input, ignore_cols)
    except ValueError as e:
        logger.error(f"⚠️  Error: {str(e)}")
        return

    if target is None:
        target = dataset.columns[-1]

    columns = [c for c in dataset.columns if c != target]

    X = dataset[columns].values
    y = dataset[target].values

    automl = AutoML(
        output=VectorCategorical(),
        search_kwargs=dict(
            evaluation_timeout=evaluation_timeout,
            memory_limit=memory_limit,
            search_timeout=search_timeout,
            pop_size=pop_size,
        ),
        random_state=random_state,
        search_iterations=iterations,
    )

    console.print(f"🏃 Training on {len(dataset)} items.")
    automl.fit(X, y, logger=RichLogger())

    with output.open("wb") as fp:
        automl.save(fp)

    console.print(f"💾 Saving model to [green]{output.absolute()}[/].")


@automl_app.command("predict")
def automl_predict(
    input: Path,
    output: Path = Path("output.csv"),
    model: Path = Path("automl.bin"),
    ignore_cols: List[int] = typer.Option([]),
    format: str = None,
):
    """
    🔮 Predict with a previously trained AutoML instance.
    """

    try:
        dataset = _load_dataset(format, input, ignore_cols)
    except ValueError as e:
        logger.error(f"⚠️  Error: {str(e)}")
        return

    try:
        with model.open("rb") as fp:
            automl = AutoML.load(fp)
    except TypeError as e:
        logger.error(f"⚠️  Error: {str(e)}")
        return

    console.print(f"🔮 Predicting {len(dataset)} items with the pipeline:")
    console.print(repr(automl.best_pipeline_))

    X = dataset.values
    y = automl.predict(X)

    with output.open("wt") as fp:
        df = pd.DataFrame(y, columns=["y"])
        df.to_csv(fp)

    console.print(f"💾 Predictions saved to [blue]{output.absolute()}[/]")


@automl_app.command("inspect")
def automl_inspect(model: Path = Path("automl.bin")):
    """
    🔍 Inspect a trained AutoML model.
    """

    with model.open("rb") as fp:
        automl = AutoML.load(fp)

    console.print(f"🔍 Inspecting AutoML model: [green]{model.absolute()}[/]")

    console.print(f"⭐ Best pipeline (score={automl.best_score_:0.3f}):")
    console.print(repr(automl.best_pipeline_))


@data_app.callback()
def data_callback():
    """
    📚 Download, inspect, and generate training data.
    """


@data_app.command("list")
def data_list():
    """
    🔍 List the available datasets.
    """

    datasets = get_datasets_list()

    table = Table("📚 Dataset", "💾", "🔗 URL")

    for item, url in sorted(datasets.items(), key=lambda t: t[0]):
        path = datapath(item)

        if path.exists():
            table.add_row(item, "✔️", url)
        else:
            table.add_row(item, "", url)

    console.print(table)


@data_app.command("download")
def data_download(
    datasets: List[str] = typer.Argument(
        ..., help="Name of one or more specific datasets to download, or 'all'."
    )
):
    """
    ⏬ Download a dataset.

    Pass a name to directly download that dataset.
    Otherwise, this command will show an interactive menu.
    """

    if "all" in datasets:
        datasets = get_datasets_list().keys()

    for dataset in datasets:
        download(dataset)


@data_app.command("gen")
def data_generate():
    """
    🎲 Generate a random dataset.
    """


if __name__ == "__main__":
    try:
        app(prog_name="autogoal")
    except Exception as e:
        console.print(f'⚠️  The command failed with message:\n"{str(e)}".')

        if console.input("❓ Do you want to inspect the traceback? \[y/N] ") == "y":
            logger.exception("Check the traceback below.")
